"""This module offers better integration with the itemadapter package."""
from types import MappingProxyType
from typing import Any, Collection, Iterator, KeysView

from itemadapter.adapter import AttrsAdapter

from zyte_common_items.base import Item


def _is_empty(value):
    """Return ``True`` if the value is to be considered empty for the purpose
    of excluding it from serialization.

    Empty values include: ``None``, empty collections (tuples, lists, etc.).

    Non-empty values include: empty ``bytes`` or ``str``, ``False``, ``0``.

    *value* is assumed not to be a mapping, which should be treated as a
    non-empty value, but this function would treat as an empty value.
    """
    return value is None or (
        not value
        and not isinstance(value, (bytes, str))
        and isinstance(value, Collection)
    )


class ZyteItemAdapter(AttrsAdapter):
    """Wrap an :ref:`item <items>` to interact with its content as if it was
    a dictionary.

    It can be :ref:`configured <configuration>` into itemadapter_ to improve
    interaction with :ref:`items <items>` for itemadapter users like Scrapy_.

    In extends AttrsAdapter_ with the following features:

    -   Allows interaction and serialization of fields from
        :attr:`~zyte_common_items.Item._unknown_fields_dict` as if they
        were regular item fields.

    -   Removes keys with empty values from the output of
        `ItemAdapter.asdict()`_, for a cleaner output.

    .. _AttrsAdapter: https://github.com/scrapy/itemadapter#built-in-adapters
    .. _itemadapter: https://github.com/scrapy/itemadapter#itemadapter
    .. _ItemAdapter.asdict(): https://github.com/scrapy/itemadapter#asdict---dict
    .. _Scrapy: https://scrapy.org/
    """

    @classmethod
    def is_item(cls, item: Any) -> bool:
        return isinstance(item, Item)

    def get_field_meta(self, field_name: str) -> MappingProxyType:
        if field_name in self._fields_dict:
            return self._fields_dict[field_name].metadata  # type: ignore
        elif field_name in self.item._unknown_fields_dict:
            return MappingProxyType({})
        raise KeyError(field_name)

    def field_names(self) -> KeysView:
        return KeysView({**self._fields_dict, **self.item._unknown_fields_dict})

    def __getitem__(self, field_name: str) -> Any:
        if field_name in self._fields_dict:
            return getattr(self.item, field_name)
        elif field_name in self.item._unknown_fields_dict:
            return self.item._unknown_fields_dict[field_name]
        raise KeyError(field_name)

    def __setitem__(self, field_name: str, value: Any) -> None:
        if field_name in self._fields_dict:
            setattr(self.item, field_name, value)
        else:
            self.item._unknown_fields_dict[field_name] = value

    def __delitem__(self, field_name: str) -> None:
        if field_name in self._fields_dict:
            del self._fields_dict[field_name]
            delattr(self.item, field_name)
        elif field_name in self.item._unknown_fields_dict:
            del self.item._unknown_fields_dict[field_name]
        else:
            raise KeyError(
                f"Object of type {self.item.__class__.__name__} does not contain a field with name {field_name}"
            )

    def __iter__(self) -> Iterator:
        fields = [
            attr
            for attr in self._fields_dict
            if not _is_empty(getattr(self.item, attr))
        ]
        fields.extend(
            attr
            for attr in self.item._unknown_fields_dict
            if not _is_empty(self.item._unknown_fields_dict[attr])
        )
        return iter(fields)


class ZyteItemKeepEmptyAdapter(ZyteItemAdapter):
    """Similar to :class:`~.ZyteItemAdapter` but doesn't remove empty values.

    It is intended to be used in tests and other use cases where it's important
    to differentiate between empty and missing fields.
    """

    def __iter__(self) -> Iterator:
        fields = [attr for attr in self._fields_dict if hasattr(self.item, attr)]
        fields.extend(self.item._unknown_fields_dict)
        return iter(fields)



import base64

import pytest

from zyte_common_items import Header, Request


def test_request_no_kwargs():
    req = Request("http://example.com")
    assert req.url == "http://example.com"
    assert req.method == "GET"
    assert req.body is None
    assert req.body_bytes is None
    assert req.headers is None


def test_request_body_bytes():
    body = base64.b64encode(b"request body").decode()
    req = Request("http://example.com", body=body, method="POST")
    assert req.method == "POST"
    assert req.body == body
    assert req.body_bytes == b"request body"


def test_request_to_scrapy_basic():
    scrapy = pytest.importorskip("scrapy")
    req = Request("http://example.com")
    scrapy_req = req.to_scrapy(None)

    assert isinstance(scrapy_req, scrapy.Request)
    assert scrapy_req.url == "http://example.com"
    assert scrapy_req.method == "GET"
    assert scrapy_req.callback is None


def test_request_to_scrapy_complex():
    scrapy = pytest.importorskip("scrapy")

    def callback():
        pass

    body = base64.b64encode(b"request body").decode()
    headers = [Header(name="Header", value="Value")]

    req = Request(
        "http://example.com",
        headers=headers,
        body=body,
        method="POST",
        name="link name",
    )

    scrapy_req = req.to_scrapy(callback=callback, meta={"foo": "bar"}, priority=5)

    assert isinstance(scrapy_req, scrapy.Request)
    assert scrapy_req.url == "http://example.com"
    assert scrapy_req.meta == {"foo": "bar"}
    assert scrapy_req.method == "POST"
    assert scrapy_req.callback is callback
    assert scrapy_req.body == b"request body"
    assert scrapy_req.headers["Header"] == b"Value"
    assert scrapy_req.priority == 5


def test_request_to_scrapy_headers_with_the_same_name():
    pytest.importorskip("scrapy")

    headers = [Header(name="name", value="value1"), Header(name="name", value="value2")]

    req = Request("http://example.com", headers=headers)
    scrapy_req = req.to_scrapy(callback=None)
    assert scrapy_req.headers.getlist("name") == [b"value1", b"value2"]



from copy import copy

import pytest

import zyte_common_items
from zyte_common_items import (
    AdditionalProperty,
    Address,
    AggregateRating,
    Amenity,
    Article,
    ArticleFromList,
    ArticleList,
    ArticleListMetadata,
    ArticleMetadata,
    ArticleNavigation,
    ArticleNavigationMetadata,
    Audio,
    Author,
    BaseSalary,
    Brand,
    Breadcrumb,
    BusinessPlace,
    BusinessPlaceMetadata,
    Gtin,
    Header,
    HiringOrganization,
    Image,
    JobLocation,
    JobPosting,
    JobPostingMetadata,
    Link,
    Metadata,
    NamedLink,
    OpeningHoursItem,
    ParentPlace,
    ProbabilityMetadata,
    ProbabilityRequest,
    Product,
    ProductFromList,
    ProductList,
    ProductListMetadata,
    ProductMetadata,
    ProductNavigation,
    ProductNavigationMetadata,
    ProductVariant,
    RealEstate,
    RealEstateArea,
    RealEstateMetadata,
    Request,
    StarRating,
    Video,
)

_ARTICLE_FROM_LIST_ALL_KWARGS: dict = {
    "articleBody": "We held the 2022 Web Data Extraction Summit three weeks ago. I wanted to extend a huge thank you to everyone who came, especially our guest speakers, who shared some great insights throughout the day.",
    "authors": [
        Author(
            name="Honorable Zytan",
            nameRaw="Honorable Zytan",
        ),
    ],
    "datePublished": "2022-10-25T00:00:00Z",
    "datePublishedRaw": "October 25, 2022",
    "headline": "Reflecting on the 2022 Web Data Extraction Summit | Zyte",
    "images": [
        Image(
            "https://www.zyte.com/wp-content/uploads/2021/10/Extract-Summit-2021_Blog-Summary_Header-85.png"
        ),
    ],
    "inLanguage": "en",
    "mainImage": Image(
        "https://www.zyte.com/wp-content/uploads/2021/10/Extract-Summit-2021_Blog-Summary_Header-85.png"
    ),
    "metadata": ProbabilityMetadata(
        probability=0.8437236322711676,
    ),
    "url": "https://www.zyte.com/blog/reflecting-on-the-2022-web-data-extraction-summit-zyte/",
}
_ARTICLE_MIN_KWARGS: dict = {
    "url": "https://www.zyte.com/blog/product-data-extraction-automatic",
}
_ARTICLE_ALL_KWARGS: dict = {
    **_ARTICLE_MIN_KWARGS,
    "headline": "Guarantee the best results for product data extraction",
    "datePublished": "2022-08-26T20:56:16Z",
    "datePublishedRaw": "August 26, 2022",
    "dateModified": "2022-08-27T21:48:55Z",
    "dateModifiedRaw": "August 25, 2022",
    "authors": [Author()],
    "breadcrumbs": [
        Breadcrumb(
            name="Homepage",
            url="https://www.zyte.com",
        ),
        Breadcrumb(
            name="Blog",
            url="https://www.zyte.com/blog/",
        ),
        Breadcrumb(
            name="Guarantee the best results for product data extraction",
        ),
    ],
    "inLanguage": "en",
    "mainImage": Image(
        "https://www.zyte.com/wp-content/uploads/2022/08/product-data-extraction-zyte.png"
    ),
    "images": [
        Image(
            "https://www.zyte.com/wp-content/uploads/2022/08/product-data-extraction-zyte.png"
        ),
        Image(
            "https://www.zyte.com/wp-content/uploads/2021/06/Webinar_Product-Data-Quality-01-1024x536.png"
        ),
        Image(
            "https://www.zyte.com/wp-content/uploads/2022/08/product-data-extraction-automatic.png"
        ),
    ],
    "description": "Product Data Extraction helps understand how consumers use a specific product, foresee improvements and adjustments that ultimately increase sales and demand.",
    "articleBody": "When businesses operate in a competitive environment it is imperative to know what their competitors are charging in real-time and this can be hard to keep track of. For any data driven organization, implementing a solution that automatically extracts product data from websites in real-time and at scale, is indispensable to stay ahead of the competition. Setting an automatic process for product data extraction can be a powerful tool for data driven businesses of all sizes. You can extract specific products information including offers, price, currency, and availability. Provided that the data extraction process is able to identify the key attributes of a product, it can then use this information to create reports with insights into the behavior of a particular product. For this to work, it is key to have the right procedures in place that guarantee the best results for your automation product extraction projects. By doing so, your organization can better understand how consumers are using a specific product, foresee any necessary adjustments to improve the user experience and as a result, increase sales and demand. In this article, we'll explain not just everything you need to know to get started, but also show you how to guarantee the best results when working with product data extraction.\n\nWhat is automatic product extraction?\n\n...",
    "articleBodyHtml": '<article>\n<p>When businesses operate in a competitive environment it is imperative to know what their competitors are charging in real-time and this can be hard to keep track of. For any data driven organization, implementing a solution that automatically extracts product data from websites in real-time and at scale, is indispensable to stay ahead of the competition.</p> <p>Setting an automatic process for <a href="https://docs.zyte.com/automatic-extraction/product.html">product data extraction</a> can be a powerful tool for data driven businesses of all sizes. You can extract specific products information including offers, price, currency, and availability. Provided that the data extraction process is able to <a href="/blog/automatic-extraction-data-extractor-review/">identify the key attributes of a product</a>, it can then use this information to create reports with insights into the behavior of a particular product.</p>\n<p>For this to work, it is key to have the right procedures in place that guarantee the best results for your automation product extraction projects. By doing so, your organization can better understand how consumers are using a specific product, foresee any necessary adjustments to improve the user experience and as a result, increase sales and demand.</p>\n<p>In this article, we\'ll explain not just everything you need to know to get started, but also show you how to guarantee the best results when working with product data extraction.</p>\n<h2><strong>What is automatic product extraction?</strong></h2>\n...\n</article>',
    "videos": [
        Video(
            "https://player.vimeo.com/video/90592864?title=0&byline=0&portrait=0&color=8dc7dc"
        ),
    ],
    "audios": [
        Audio("https://example.com/audiofile.mp3"),
    ],
    "canonicalUrl": "https://www.zyte.com/blog/product-data-extraction-automatic/",
    "url": "https://www.zyte.com/blog/product-data-extraction-automatic",
    "metadata": ArticleMetadata(
        dateDownloaded="2022-12-31T13:01:54Z",
        probability=1.0,
    ),
}
_ARTICLE_LIST_MIN_KWARGS: dict = {
    "url": "https://www.zyte.com/blog/extract-summit-blog/",
}
_ARTICLE_LIST_ALL_KWARGS: dict = {
    **_ARTICLE_LIST_MIN_KWARGS,
    "canonicalUrl": "https://www.zyte.com/blog/extract-summit-blog/",
    "articles": [
        ArticleFromList(
            url="https://www.zyte.com/blog/reflecting-on-the-2022-web-data-extraction-summit-zyte/",
        ),
    ],
    "breadcrumbs": [
        Breadcrumb(
            name="Homepage",
            url="https://www.zyte.com/",
        ),
        Breadcrumb(
            name="Blog",
            url="https://www.zyte.com/blog/",
        ),
        Breadcrumb(
            name="Extract Summit",
            url="https://www.zyte.com/blog/extract-summit-blog/",
        ),
    ],
    "metadata": ArticleListMetadata(
        dateDownloaded="2022-12-31T13:01:54Z",
    ),
}

_ARTICLE_NAVIGATION_MIN_KWARGS: dict = {"url": "https://example.com/real-estate/12345"}

_ARTICLE_NAVIGATION_ALL_KWARGS: dict = {
    **_ARTICLE_NAVIGATION_MIN_KWARGS,
    "categoryName": "Swiss Watches",
    "subCategories": [
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/category/books/",
            method="POST",
            body="YmFzZTY0LWVuY29kZWQ=",
            headers=[Header(name="content-type", value="text/json")],
            name="Travel",
            metadata=ProbabilityMetadata(probability=0.99),
        ),
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/category/books/mystery_3/index.html",
            name="Mystery",
            metadata=ProbabilityMetadata(probability=0.97),
        ),
    ],
    "items": [
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/in-her-wake_980",
            method="POST",
            body="YmFzZTY0LWVuY29kZWQ=",
            headers=[Header(name="content-type", value="text/json")],
            name="In Her Wake",
            metadata=ProbabilityMetadata(probability=0.99),
        ),
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/how-music-works_979/index.html",
            name="How Music Works",
            metadata=ProbabilityMetadata(probability=0.98),
        ),
    ],
    "nextPage": Request(
        url="http://books.toscrape.com/catalogue/page-3.html",
        name="Next page",
        method="POST",
        body="Im9rIg==",
        headers=[Header(name="content-type", value="text/json")],
    ),
    "pageNumber": 5,
    "metadata": ArticleNavigationMetadata(dateDownloaded="2022-12-31T13:01:54Z"),
}

_BUSINESS_PLACE_MIN_KWARGS: dict = {}

_BUSINESS_PLACE_ALL_KWARGS: dict = {
    **_BUSINESS_PLACE_MIN_KWARGS,
    "url": "https://place-listing.example/place/12345",
    "placeId": "ChIJ02PI3bsaB3wRh-IwsWeJ0nI",
    "name": "Sheraton Kauai Resort",
    "actions": [
        NamedLink(name="Reserve Table", url="https://example.com"),
        NamedLink(name="Book Room", url="https://example.com"),
    ],
    "additionalProperties": [
        AdditionalProperty(name="Popular for", value="Lunch,Dinner")
    ],
    "address": Address(addressRaw="2440 Hoonani Rd, Koloa, HI 96756, US"),
    "reservationAction": NamedLink(name="Reserve a table", url="https://example.com"),
    "categories": ["Hotel", "Wedding venue"],
    "description": "Oceanfront resort offering Hawaiian dining, plus 2 outdoor pools & a fitness center.",
    "features": ["Pool", "Free Parking"],
    "map": "https://map.example/place/12345",
    "images": [
        Image("http://example.com/image1.png"),
    ],
    "amenityFeatures": [Amenity(name="Free WiFi", value=True)],
    "aggregateRating": AggregateRating(
        bestRating=5.0,
        ratingValue=2.5,
        reviewCount=123,
    ),
    "starRating": StarRating(raw="4-star Hotel"),
    "containedInPlace": ParentPlace(name="Gateway Plaza", placeId="ChIJCVDfcE_GBtq"),
    "openingHours": [OpeningHoursItem(dayOfWeek="Monday", opens="10:00")],
    "reviewSites": [NamedLink(name="Expedia", url="https://example.com")],
    "telephone": "(808) 742-1661",
    "priceRange": "zz",
    "timezone": "Pacific/Honolulu",
    "isVerified": True,
    "website": "https://example.com",
    "tags": ["Dogs", "Family"],
    "metadata": BusinessPlaceMetadata(
        searchText="hotel", dateDownloaded="2022-12-31T13L01L54Z", probability=0.95
    ),
}

_PRODUCT_FROM_LIST_ALL_KWARGS: dict = {
    "currency": "USD",
    "currencyRaw": "$",
    "mainImage": Image("http://example.com/image1.png"),
    "metadata": ProbabilityMetadata(
        probability=1.0,
    ),
    "name": "White two-door refrigerator",
    "price": "9999.99",
    "productId": "A123DK9823",
    "regularPrice": "11999.99",
    "url": "https://example.com/?product=product22",
}
_PRODUCT_VARIANT_ALL_KWARGS: dict = {
    "additionalProperties": [AdditionalProperty("foo", "bar")],
    "availability": "InStock",
    "canonicalUrl": "https://example.com/product22",
    "color": "white",
    "currency": "USD",
    "currencyRaw": "$",
    "gtin": [Gtin("foo", "bar")],
    "images": [
        Image("http://example.com/image1.png"),
    ],
    "mainImage": Image("http://example.com/image1.png"),
    "mpn": "HSC0424PP",
    "name": "White two-door refrigerator",
    "price": "9999.99",
    "productId": "A123DK9823",
    "regularPrice": "11999.99",
    "size": "XL",
    "sku": "A123DK9823",
    "style": "polka dots",
    "url": "https://example.com/?product=product22",
}
_PRODUCT_MIN_KWARGS: dict = {
    "url": "https://example.com/?product=product22",
}
_PRODUCT_ALL_KWARGS: dict = {
    **_PRODUCT_MIN_KWARGS,
    **_PRODUCT_VARIANT_ALL_KWARGS,
    "aggregateRating": AggregateRating(
        bestRating=5.0,
        ratingValue=2.5,
        reviewCount=123,
    ),
    "brand": Brand("Ka-pow"),
    "breadcrumbs": [
        Breadcrumb(name="Level 1", url="http://example.com/level1"),
        Breadcrumb(name="Level 2", url="http://example.com/level1/level2"),
    ],
    "description": (
        "Full freshness all over the fridge\n5 Conversion Modes on demand\nSuper Cooling Plus™"
    ),
    "descriptionHtml": (
        "<article><p>Full freshness all over the fridge</p>"
        "<p>5 Conversion Modes on demand</p>"
        "<p>Super Cooling Plus&trade;</p></article>"
    ),
    "features": ["Easily store fragile products.", "Bluetooth connectivity."],
    "metadata": ProductMetadata(
        dateDownloaded="2022-12-31T13:01:54Z",
        probability=1.0,
    ),
    "variants": [ProductVariant()],
}
_PRODUCT_LIST_MIN_KWARGS: dict = {
    "url": "https://example.com/swiss-watches?sort=new-first",
}
_PRODUCT_LIST_ALL_KWARGS: dict = {
    **_PRODUCT_LIST_MIN_KWARGS,
    "breadcrumbs": [
        Breadcrumb(name="Level 1", url="http://example.com/level1"),
        Breadcrumb(name="Level 2", url="http://example.com/level1/level2"),
    ],
    "canonicalUrl": "https://example.com/swiss-watches",
    "categoryName": "Swiss Watches",
    "metadata": ProductListMetadata(
        dateDownloaded="2022-12-31T13:01:54Z",
    ),
    "pageNumber": 1,
    "paginationNext": Link(
        text="foo",
        url="https://example.com/swiss-watches?sort=new-first&page=2",
    ),
    "products": [ProductFromList()],
}

_REAL_ESTATE_MIN_KWARGS: dict = {
    "url": "https://example.com/real-estate/12345",
}

_REAL_ESTATE_ALL_KWARGS: dict = {
    **_REAL_ESTATE_MIN_KWARGS,
    "breadcrumbs": [
        Breadcrumb(name="Level 1", url="http://example.com/level1"),
        Breadcrumb(name="Level 2", url="http://example.com/level1/level2"),
    ],
    "realEstateId": "12345",
    "name": "Beautiful 2 bedroom apartment",
    "datePublished": "2022-12-31T13:01:54Z",
    "datePublishedRaw": "2022-12-31",
    "description": "Beautiful 2 bedroom apartment in the heart of the city",
    "mainImage": Image("http://example.com/image1.png"),
    "images": [
        Image("http://example.com/image1.png"),
        Image("http://example.com/image2.png"),
    ],
    "address": Address(addressRaw="2440 Hoonani Rd, Koloa, HI 96756, US"),
    "area": RealEstateArea(
        value=100,
        unitCode="SQMT",
        areaType="LOT",
        raw="100 sqm",
    ),
    "numberOfBathroomsTotal": 2,
    "numberOfFullBathrooms": 1,
    "numberOfPartialBathrooms": 1,
    "numberOfBedrooms": 2,
    "numberOfRooms": 2,
    "tradeType": "RENT",
    "price": "456456",
    "rentalPeriod": "MONTH",
    "currency": "USD",
    "currencyRaw": "$",
    "additionalProperties": [
        AdditionalProperty(name="Parking", value="Garage"),
        AdditionalProperty(name="Pets", value="Allowed"),
    ],
    "propertyType": "flat",
    "yearBuilt": 1997,
    "virtualTourUrl": "https://example.com",
    "metadata": RealEstateMetadata(
        dateDownloaded="2022-12-31T13:01:54Z",
        probability=1.0,
    ),
}

_PRODUCT_NAVIGATION_MIN_KWARGS: dict = {"url": "https://example.com/real-estate/12345"}

_PRODUCT_NAVIGATION_ALL_KWARGS: dict = {
    **_PRODUCT_NAVIGATION_MIN_KWARGS,
    "categoryName": "Swiss Watches",
    "subCategories": [
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/category/books/",
            method="POST",
            body="YmFzZTY0LWVuY29kZWQ=",
            headers=[Header(name="content-type", value="text/json")],
            name="Travel",
            metadata=ProbabilityMetadata(probability=0.99),
        ),
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/category/books/mystery_3/index.html",
            name="Mystery",
            metadata=ProbabilityMetadata(probability=0.97),
        ),
    ],
    "items": [
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/in-her-wake_980",
            method="POST",
            body="YmFzZTY0LWVuY29kZWQ=",
            headers=[Header(name="content-type", value="text/json")],
            name="In Her Wake",
            metadata=ProbabilityMetadata(probability=0.99),
        ),
        ProbabilityRequest(
            url="http://books.toscrape.com/catalogue/how-music-works_979/index.html",
            name="How Music Works",
            metadata=ProbabilityMetadata(probability=0.98),
        ),
    ],
    "nextPage": Request(
        url="http://books.toscrape.com/catalogue/page-3.html",
        name="Next page",
        method="POST",
        body="Im9rIg==",
        headers=[Header(name="content-type", value="text/json")],
    ),
    "pageNumber": 5,
    "metadata": ProductNavigationMetadata(dateDownloaded="2022-12-31T13:01:54Z"),
}

_JOB_POSTING_MIN_KWARGS: dict = {
    "url": "https://example.com/viewjob/12345",
}

_JOB_POSTING_ALL_KWARGS: dict = {
    **_JOB_POSTING_MIN_KWARGS,
    "jobPostingId": "12345",
    "datePublished": "2019-06-19T00:00:00-05:00",
    "datePublishedRaw": "19 June 2019",
    "dateModified": "2019-06-21T00:00:00-05:00",
    "dateModifiedRaw": "21 June 2019",
    "validThrough": "2019-07-19T00:00:00-05:00",
    "validThroughRaw": "19 July 2019",
    "jobTitle": "Software Engineer",
    "headline": "Are you our next Software Engineer?",
    "jobLocation": JobLocation(raw="New York, NY"),
    "description": "We are looking for a Software Engineer to join our team."
    "- 35 days holiday"
    "- 15% bonus"
    "- flexible working arrangements",
    "descriptionHtml": "<p>We are looking for a Software Engineer to join our team.</p>"
    "<ul>"
    "<li>35 days holiday</li>"
    "<li>15% bonus</li>"
    "<li>flexible working arrangements</li>"
    "</ul>",
    "employmentType": "Full-time",
    "baseSalary": BaseSalary(
        raw="$53,000-$55,000 a year",
        valueMin="53000",
        valueMax="55000",
        rateType="yearly",
        currencyRaw="$",
        currency="USD",
    ),
    "requirements": [
        "Experience in managing diverse teams",
        "Great sense of responsibility",
        "5+ years of proven experience in sales",
        "Ability to travel",
    ],
    "hiringOrganization": HiringOrganization(
        name="ACME Corp.", nameRaw="ACME Corp., US", id="54321"
    ),
    "jobStartDate": "2019-08-01T00:00:00-05:00",
    "jobStartDateRaw": "01 August 2019",
    "remoteStatus": "Remote",
    "metadata": JobPostingMetadata(
        dateDownloaded="2022-12-31T13:01:54Z",
        probability=0.95,
        searchText="Software Engineer",
    ),
}


def test_article_all_fields():
    article = Article(**_ARTICLE_ALL_KWARGS)
    for field in list(_ARTICLE_ALL_KWARGS):
        assert getattr(article, field) == _ARTICLE_ALL_KWARGS[field]


def test_article_min_fields():
    article = Article(**_ARTICLE_MIN_KWARGS)
    for field in list(_ARTICLE_ALL_KWARGS):
        if field in _ARTICLE_MIN_KWARGS:
            continue
        assert getattr(article, field) is None


def test_article_missing_fields():
    for required_field in list(_ARTICLE_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_ARTICLE_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            Article(**incomplete_kwargs)


def test_article_list_all_fields():
    article_list = ArticleList(**_ARTICLE_LIST_ALL_KWARGS)
    for field in list(_ARTICLE_LIST_ALL_KWARGS):
        assert getattr(article_list, field) == _ARTICLE_LIST_ALL_KWARGS[field]


def test_article_list_min_fields():
    article_list = ArticleList(**_ARTICLE_LIST_MIN_KWARGS)
    for field in list(_ARTICLE_LIST_MIN_KWARGS):
        if field in _ARTICLE_LIST_MIN_KWARGS:
            continue
        assert getattr(article_list, field) is None


def test_article_list_missing_fields():
    for required_field in list(_ARTICLE_LIST_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_ARTICLE_LIST_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            ArticleList(**incomplete_kwargs)


def test_article_from_list_all_fields():
    article_from_list = ArticleFromList(**_ARTICLE_FROM_LIST_ALL_KWARGS)
    for field in list(_ARTICLE_FROM_LIST_ALL_KWARGS):
        assert getattr(article_from_list, field) == _ARTICLE_FROM_LIST_ALL_KWARGS[field]


def test_article_from_list_min_fields():
    article_from_list = ArticleFromList()
    for field in list(_ARTICLE_FROM_LIST_ALL_KWARGS):
        assert getattr(article_from_list, field) is None


def test_article_navigation_all_fields():
    article_navigation = ArticleNavigation(**_ARTICLE_NAVIGATION_ALL_KWARGS)
    for field in list(_ARTICLE_NAVIGATION_ALL_KWARGS):
        assert (
            getattr(article_navigation, field) == _ARTICLE_NAVIGATION_ALL_KWARGS[field]
        )


def test_article_navigation_min_fields():
    article_navigation = ArticleNavigation(**_ARTICLE_NAVIGATION_MIN_KWARGS)
    for field in list(_ARTICLE_NAVIGATION_ALL_KWARGS):
        if field in _ARTICLE_NAVIGATION_MIN_KWARGS:
            continue
        assert getattr(article_navigation, field) is None


def test_article_navigation_missing_fields():
    for required_field in list(_ARTICLE_NAVIGATION_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_ARTICLE_NAVIGATION_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            ArticleNavigation(**incomplete_kwargs)


def test_product_all_fields():
    product = Product(**_PRODUCT_ALL_KWARGS)
    for field in list(_PRODUCT_ALL_KWARGS):
        assert getattr(product, field) == _PRODUCT_ALL_KWARGS[field]


def test_product_min_fields():
    product = Product(**_PRODUCT_MIN_KWARGS)
    for field in list(_PRODUCT_ALL_KWARGS):
        if field in _PRODUCT_MIN_KWARGS:
            continue
        assert getattr(product, field) is None


def test_product_missing_fields():
    for required_field in list(_PRODUCT_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_PRODUCT_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            Product(**incomplete_kwargs)


def test_product_list_all_fields():
    product_list = ProductList(**_PRODUCT_LIST_ALL_KWARGS)
    for field in list(_PRODUCT_LIST_ALL_KWARGS):
        assert getattr(product_list, field) == _PRODUCT_LIST_ALL_KWARGS[field]


def test_product_list_min_fields():
    product_list = ProductList(**_PRODUCT_LIST_MIN_KWARGS)
    for field in list(_PRODUCT_LIST_MIN_KWARGS):
        if field in _PRODUCT_LIST_MIN_KWARGS:
            continue
        assert getattr(product_list, field) is None


def test_product_list_missing_fields():
    for required_field in list(_PRODUCT_LIST_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_PRODUCT_LIST_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            ProductList(**incomplete_kwargs)


def test_product_from_list_all_fields():
    product_from_list = ProductFromList(**_PRODUCT_FROM_LIST_ALL_KWARGS)
    for field in list(_PRODUCT_FROM_LIST_ALL_KWARGS):
        assert getattr(product_from_list, field) == _PRODUCT_FROM_LIST_ALL_KWARGS[field]


def test_product_from_list_min_fields():
    product_from_list = ProductFromList()
    for field in list(_PRODUCT_FROM_LIST_ALL_KWARGS):
        assert getattr(product_from_list, field) is None


def test_product_variant_all_fields():
    product_variant = ProductVariant(**_PRODUCT_VARIANT_ALL_KWARGS)
    for field in list(_PRODUCT_VARIANT_ALL_KWARGS):
        assert getattr(product_variant, field) == _PRODUCT_VARIANT_ALL_KWARGS[field]


def test_product_variant_min_fields():
    product_variant = ProductVariant()
    for field in list(_PRODUCT_VARIANT_ALL_KWARGS):
        assert getattr(product_variant, field) is None


def test_business_place_all_fields():
    place = BusinessPlace(**_BUSINESS_PLACE_ALL_KWARGS)
    for field in list(_BUSINESS_PLACE_ALL_KWARGS):
        assert getattr(place, field) == _BUSINESS_PLACE_ALL_KWARGS[field]


def test_business_place_min_fields():
    place = BusinessPlace(**_BUSINESS_PLACE_MIN_KWARGS)
    for field in list(_BUSINESS_PLACE_ALL_KWARGS):
        if field in _BUSINESS_PLACE_MIN_KWARGS:
            continue
        assert getattr(place, field) is None


def test_business_place_missing_fields():
    for required_field in list(_BUSINESS_PLACE_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_BUSINESS_PLACE_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            BusinessPlace(**incomplete_kwargs)


def test_real_estate_all_fields():
    real_estate = RealEstate(**_REAL_ESTATE_ALL_KWARGS)
    for field in list(_REAL_ESTATE_ALL_KWARGS):
        assert getattr(real_estate, field) == _REAL_ESTATE_ALL_KWARGS[field]


def test_real_estate_min_fields():
    real_estate = RealEstate(**_REAL_ESTATE_MIN_KWARGS)
    for field in list(_REAL_ESTATE_ALL_KWARGS):
        if field in _REAL_ESTATE_MIN_KWARGS:
            continue
        assert getattr(real_estate, field) is None


def test_real_estate_missing_fields():
    for required_field in list(_REAL_ESTATE_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_REAL_ESTATE_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            RealEstate(**incomplete_kwargs)


def test_product_navigation_all_fields():
    product_navigation = ProductNavigation(**_PRODUCT_NAVIGATION_ALL_KWARGS)
    for field in list(_PRODUCT_NAVIGATION_ALL_KWARGS):
        assert (
            getattr(product_navigation, field) == _PRODUCT_NAVIGATION_ALL_KWARGS[field]
        )


def test_product_navigation_min_fields():
    product_navigation = ProductNavigation(**_PRODUCT_NAVIGATION_MIN_KWARGS)
    for field in list(_PRODUCT_NAVIGATION_ALL_KWARGS):
        if field in _PRODUCT_NAVIGATION_MIN_KWARGS:
            continue
        assert getattr(product_navigation, field) is None


def test_product_navigation_missing_fields():
    for required_field in list(_PRODUCT_NAVIGATION_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_PRODUCT_NAVIGATION_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            ProductNavigation(**incomplete_kwargs)


def test_metadata():
    """Test metadata expectations for items.

    For every item:

    -   There must be a matching metadata class.

    -   The metadata attribute must be of that metadata class.

    -   If the generic Metadata object is assigned as input, it gets converted
        into an object of the right, more specific metadata class.
    """
    item_names = {
        obj_name[:-4]
        for obj_name in zyte_common_items.__dict__
        if (
            not obj_name.startswith("Base")
            and obj_name.endswith("Page")
            and obj_name != "Page"
        )
    }
    for item_name in item_names:
        cls = zyte_common_items.__dict__[item_name]
        metadata_cls = zyte_common_items.__dict__[f"{item_name}Metadata"]

        obj1 = cls.from_dict(
            {"url": "https://example.com", "metadata": {"dateDownloaded": "foo"}}
        )
        assert type(obj1.metadata) == metadata_cls
        assert obj1.metadata.dateDownloaded == "foo"

        obj2 = cls(url="https://example.com", metadata=Metadata(dateDownloaded="foo"))
        assert type(obj2.metadata) == metadata_cls
        assert obj2.metadata.dateDownloaded == "foo"
        obj2.metadata = Metadata(dateDownloaded="foo")
        assert type(obj2.metadata) == metadata_cls
        assert obj2.metadata.dateDownloaded == "foo"


def test_request():
    """Test request class conversion on assignment"""

    product_navigation = ProductNavigation(**_PRODUCT_NAVIGATION_MIN_KWARGS)
    product_navigation.subCategories = [
        Request(  # type: ignore[list-item]
            url="http://books.toscrape.com/catalogue/category/books/mystery_3/index.html",
            name="Mystery",
        ),
    ]
    assert product_navigation.subCategories[0] == ProbabilityRequest(
        url="http://books.toscrape.com/catalogue/category/books/mystery_3/index.html",
        name="Mystery",
        metadata=ProbabilityMetadata(probability=1.0),
    )


def test_job_posting_all_fields():
    job_posting = JobPosting(**_JOB_POSTING_ALL_KWARGS)
    for field in list(_JOB_POSTING_ALL_KWARGS):
        assert getattr(job_posting, field) == _JOB_POSTING_ALL_KWARGS[field]


def test_job_posting_min_fields():
    job_posting = JobPosting(**_JOB_POSTING_MIN_KWARGS)
    for field in list(_JOB_POSTING_ALL_KWARGS):
        if field in _JOB_POSTING_MIN_KWARGS:
            continue
        assert getattr(job_posting, field) is None


def test_job_posting_missing_fields():
    for required_field in list(_JOB_POSTING_MIN_KWARGS):
        incomplete_kwargs: dict = copy(_JOB_POSTING_MIN_KWARGS)
        del incomplete_kwargs[required_field]
        with pytest.raises(TypeError):
            JobPosting(**incomplete_kwargs)


@pytest.mark.parametrize(
    "cls,has_proba",
    (
        (Article, True),
        (ArticleFromList, True),
        (ArticleList, False),
        (ArticleNavigation, False),
        (BusinessPlace, True),
        (JobPosting, True),
        (Product, True),
        (ProductFromList, True),
        (ProductList, False),
        (ProductNavigation, False),
        (ProductVariant, False),
        (RealEstate, True),
    ),
)
def test_get_probability_request(cls, has_proba):
    data = {"url": "https://example.com"}

    item = cls.from_dict(data)
    assert item.get_probability() is None

    item = cls.from_dict({**data, "metadata": {"probability": 0.5}})
    if has_proba:
        assert item.get_probability() == 0.5
    else:
        assert item.get_probability() is None

    item = cls.from_dict({**data, "metadata": {"probability": 0.0}})
    if has_proba:
        assert item.get_probability() == 0.0
    else:
        assert item.get_probability() is None



# flake8: noqa
from .adapter import ZyteItemAdapter, ZyteItemKeepEmptyAdapter
from .base import Item, is_data_container
from .components import (
    AdditionalProperty,
    Address,
    AggregateRating,
    Amenity,
    ArticleListMetadata,
    ArticleMetadata,
    ArticleNavigationMetadata,
    Audio,
    Author,
    BaseSalary,
    Brand,
    Breadcrumb,
    BusinessPlaceMetadata,
    Gtin,
    Header,
    HiringOrganization,
    Image,
    JobLocation,
    JobPostingMetadata,
    Link,
    Metadata,
    NamedLink,
    OpeningHoursItem,
    ParentPlace,
    ProbabilityMetadata,
    ProbabilityRequest,
    ProductListMetadata,
    ProductMetadata,
    ProductNavigationMetadata,
    RealEstateArea,
    RealEstateMetadata,
    Request,
    StarRating,
    Video,
)
from .items import (
    Article,
    ArticleFromList,
    ArticleList,
    ArticleNavigation,
    BusinessPlace,
    JobPosting,
    Product,
    ProductFromList,
    ProductList,
    ProductNavigation,
    ProductVariant,
    RealEstate,
)
from .pages import (
    ArticleListPage,
    ArticleNavigationPage,
    ArticlePage,
    BaseArticleListPage,
    BaseArticleNavigationPage,
    BaseArticlePage,
    BaseBusinessPlacePage,
    BaseJobPostingPage,
    BasePage,
    BaseProductListPage,
    BaseProductNavigationPage,
    BaseProductPage,
    BaseRealEstatePage,
    BusinessPlacePage,
    HasMetadata,
    JobPostingPage,
    MetadataT,
    Page,
    ProductListPage,
    ProductNavigationPage,
    ProductPage,
    RealEstatePage,
)



"""Classes for data nested within items."""
import base64
from typing import List, Optional, Type

import attrs

from zyte_common_items.base import Item, ProbabilityMixin
from zyte_common_items.util import convert_to_class, url_to_str

# Metadata ####################################################################


@attrs.define(kw_only=True)
class ProbabilityMetadata(Item):
    """Data extraction process metadata that indicates a probability."""

    #: The probability (0 for 0%, 1 for 100%) that the resource features the
    #: expected data type.
    #:
    #: For example, if the extraction of a product from a given URL is
    #: requested, and that URL points to the webpage of a product with complete
    #: certainty, the value should be `1`. If with complete certainty the
    #: webpage features a job listing instead of a product, the value should be
    #: `0`. When there is no complete certainty, the value could be anything in
    #: between (e.g. `0.96`).
    probability: Optional[float] = 1.0


@attrs.define(kw_only=True)
class _ListMetadata(Item):
    """Data extraction process metadata that indicates the download date.

    See
    :class:`ArticleList.metadata <zyte_common_items.ArticleList.metadata>`.
    """

    #: Date and time when the product data was downloaded, in UTC timezone and
    #: the following format: ``YYYY-MM-DDThh:mm:ssZ``.
    dateDownloaded: Optional[str] = None


@attrs.define(kw_only=True)
class _DetailsMetadata(_ListMetadata):
    """Data extraction process metadata that indicates the download date and a
    probability."""

    #: The probability (0 for 0%, 1 for 100%) that the resource features the
    #: expected data type.
    #:
    #: For example, if the extraction of a product from a given URL is
    #: requested, and that URL points to the webpage of a product with complete
    #: certainty, the value should be `1`. If with complete certainty the
    #: webpage features a job listing instead of a product, the value should be
    #: `0`. When there is no complete certainty, the value could be anything in
    #: between (e.g. `0.96`).
    probability: Optional[float] = 1.0


@attrs.define(kw_only=True)
class Metadata(_DetailsMetadata):
    """Generic metadata class.

    It defines all attributes of metadata classes for specific item types, so
    that it can be used during extraction instead of a more specific class, and
    later converted to the corresponding, more specific metadata class.
    """

    #: The search text used to find the item.
    searchText: Optional[str] = None


@attrs.define(kw_only=True)
class ArticleMetadata(_DetailsMetadata):
    pass


@attrs.define(kw_only=True)
class ArticleListMetadata(_ListMetadata):
    pass


@attrs.define(kw_only=True)
class ArticleNavigationMetadata(_ListMetadata):
    pass


@attrs.define(kw_only=True)
class BusinessPlaceMetadata(Metadata):
    pass


@attrs.define(kw_only=True)
class JobPostingMetadata(Metadata):
    """Metadata associated with a job posting."""

    pass


@attrs.define(kw_only=True)
class ProductMetadata(_DetailsMetadata):
    pass


@attrs.define(kw_only=True)
class ProductListMetadata(_ListMetadata):
    pass


@attrs.define(kw_only=True)
class ProductNavigationMetadata(_ListMetadata):
    pass


@attrs.define(kw_only=True)
class RealEstateMetadata(_DetailsMetadata):
    pass


###############################################################################


@attrs.define
class _Media(Item):
    #: URL.
    #:
    #: When multiple URLs exist for a given media element, pointing to
    #: different-quality versions, the highest-quality URL should be used.
    #:
    #: `Data URIs`_ are not allowed in this attribute.
    #:
    #: .. _Data URIs: https://en.wikipedia.org/wiki/Data_URI_scheme
    url: str = attrs.field(converter=url_to_str)


@attrs.define
class AdditionalProperty(Item):
    """A name-value pair.

    See :attr:`Product.additionalProperties
    <zyte_common_items.Product.additionalProperties>`.
    """

    #: Name.
    name: str

    #: Value.
    value: str


@attrs.define(kw_only=True)
class AggregateRating(Item):
    """Aggregate data about reviews and ratings.

    At least one of :attr:`ratingValue` or :attr:`reviewCount` is required.

    See :attr:`Product.aggregateRating
    <zyte_common_items.Product.aggregateRating>`.
    """

    #: Maximum value of the rating system.
    bestRating: Optional[float] = None

    #: Average value of all ratings.
    ratingValue: Optional[float] = None

    #: Review count.
    reviewCount: Optional[int] = None


@attrs.define
class Audio(_Media):
    """Audio.

    See :class:`Article.audios <zyte_common_items.Article.audios>`.
    """


@attrs.define(kw_only=True)
class Author(Item):
    """Author of an article.

    See :attr:`Article.authors <zyte_common_items.Article.authors>`.
    """

    #: Email.
    email: Optional[str] = None

    #: URL of the details page of the author.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: Full name.
    name: Optional[str] = None

    #: Text from which :attr:`~zyte_common_items.Author.name` was
    #: extracted.
    nameRaw: Optional[str] = None


@attrs.define
class Brand(Item):
    """Brand.

    See :attr:`Product.brand <zyte_common_items.Product.brand>`.
    """

    #: Name as it appears on the source webpage (no post-processing).
    name: str


@attrs.define(kw_only=True)
class Breadcrumb(Item):
    """A breadcrumb from the `breadcrumb trail`_ of a webpage.

    See :attr:`Product.breadcrumbs
    <zyte_common_items.Product.breadcrumbs>`.

    .. _breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    """

    #: Displayed name.
    name: Optional[str] = None

    #: Target URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define
class Gtin(Item):
    """GTIN_ type-value pair.

    See :class:`Product.gtin <zyte_common_items.Product.gtin>`.

    .. _GTIN: https://en.wikipedia.org/wiki/Global_Trade_Item_Number
    """

    #: Identifier of the GTIN format of ``value``.
    #:
    #: One of: ``"gtin13"``, ``"gtin8"``, ``"gtin14"``, ``"isbn10"``,
    #: ``"isbn13"``, ``"ismn"``, ``"issn"``, ``"upc"``.
    type: str

    #: Value.
    #:
    #: It should only contain digits.
    value: str


@attrs.define
class Image(_Media):
    """Image.

    See for example
    :class:`Product.images <zyte_common_items.Product.images>` and
    :class:`Product.mainImage <zyte_common_items.Product.mainImage>`.
    """


@attrs.define(kw_only=True)
class Link(Item):
    """A link from a webpage to another webpage."""

    #: Displayed text.
    text: Optional[str] = None

    #: Target URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(kw_only=True)
class NamedLink(Item):
    """A link from a webpage to another webpage."""

    #: The name of the link.
    name: Optional[str] = None

    #: Target URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(kw_only=True)
class Address(Item):
    """Address item."""

    #: The raw address information, as it appears on the website.
    addressRaw: Optional[str] = None

    #: The street address of the place.
    streetAddress: Optional[str] = None

    #: The city the place is located in.
    addressCity: Optional[str] = None

    #: The locality to which the place belongs.
    addressLocality: Optional[str] = None

    #: The region of the place.
    addressRegion: Optional[str] = None

    #: The country the place is located in.
    #:
    #: The country name or the `ISO 3166-1 alpha-2 country code
    #: <https://en.wikipedia.org/wiki/ISO_3166-1>`__.
    addressCountry: Optional[str] = None

    #: The postal code of the address.
    postalCode: Optional[str] = None

    #: The auxiliary part of the postal code.
    #:
    #: It may include a state abbreviation or town name, depending on local standards.
    postalCodeAux: Optional[str] = None

    #: Geographical latitude of the place.
    latitude: Optional[float] = None

    #: Geographical longitude of the place.
    longitude: Optional[float] = None


@attrs.define(kw_only=True)
class Amenity(Item):
    """An amenity that a business place has"""

    #: Name of amenity.
    name: str

    #: Availability of the amenity.
    value: bool


@attrs.define(kw_only=True)
class StarRating(Item):
    """Official star rating of a place."""

    #: Star rating of the place, as it appears on the page, without processing.
    raw: Optional[str] = None

    #: Star rating value of the place.
    ratingValue: Optional[float] = None


@attrs.define(kw_only=True)
class ParentPlace(Item):
    """If the place is located inside another place, these are the details of the parent place."""

    #: Name of the parent place.
    name: str

    #: Identifier of the parent place.
    placeId: str


@attrs.define(kw_only=True)
class OpeningHoursItem(Item):
    """Specification of opening hours of a business place."""

    #: English weekday name.
    dayOfWeek: Optional[str] = None

    #: Opening time in ISO 8601 format, local time.
    opens: Optional[str] = None

    #: Closing time in ISO 8601 format, local time.
    closes: Optional[str] = None

    #: Day of the week, as it appears on the page, without processing.
    rawDayOfWeek: Optional[str] = None

    #: Opening time, as it appears on the page, without processing.
    rawOpens: Optional[str] = None

    #: Closing time, as it appears on the page, without processing.
    rawCloses: Optional[str] = None


@attrs.define(kw_only=True)
class RealEstateArea(Item):
    """Area of a place, with type, units, value and raw value."""

    #: Area
    value: float

    #: Unit of the value field, one of: SQMT (square meters), SQFT (square
    #: feet), ACRE (acres).
    unitCode: str

    #: Type of area, one of: LOT, FLOOR
    areaType: Optional[str] = None

    #: Area in the raw format, as it appears on the website.
    raw: str


@attrs.define(kw_only=True)
class Header(Item):
    """An HTTP header"""

    #: Name of the header
    name: str

    #: Value of the header
    value: str


@attrs.define(slots=False)
class Request(Item):
    """Describe a web request to load a page"""

    #: HTTP URL
    url: str = attrs.field(converter=url_to_str)

    #: HTTP method
    method: str = "GET"

    #: HTTP request body, Base64-encoded
    body: Optional[str] = None

    #: HTTP headers
    headers: Optional[List[Header]] = None

    #: Name of the page being requested.
    name: Optional[str] = None

    _body_bytes = None

    @property
    def body_bytes(self) -> Optional[bytes]:
        """Request.body as bytes"""
        # todo: allow to set body bytes in __init__, to avoid encoding/decoding.
        if self._body_bytes is None:
            if self.body is not None:
                self._body_bytes = base64.b64decode(self.body)
        return self._body_bytes

    def to_scrapy(self, callback, **kwargs):
        """
        Convert a request to scrapy.Request.
        All kwargs are passed to scrapy.Request as-is.
        """
        import scrapy

        header_list = [(header.name, header.value) for header in self.headers or []]

        return scrapy.Request(
            url=self.url,
            callback=callback,
            method=self.method or "GET",
            headers=header_list,
            body=self.body_bytes,
            **kwargs,
        )


@attrs.define
class Video(_Media):
    """Video.

    See :class:`Article.videos <zyte_common_items.Article.videos>`.
    """


def cast_request(value: Request, cls: Type[Request]) -> Request:
    new_value = convert_to_class(value, cls)
    if type(value) is Request and cls is ProbabilityRequest:
        new_value.metadata = ProbabilityMetadata(probability=1.0)
    return new_value


def request_list_processor(request_list):
    return [cast_request(request, ProbabilityRequest) for request in request_list]


@attrs.define(kw_only=True)
class ProbabilityRequest(Request, ProbabilityMixin):
    """A :class:`Request` that includes a probability value."""

    #: Data extraction process metadata.
    metadata: Optional[ProbabilityMetadata] = None


@attrs.define(kw_only=True)
class JobLocation(Item):
    """Location of a job offer."""

    #: Job location, as it appears on the website.
    raw: Optional[str] = None


@attrs.define(kw_only=True)
class BaseSalary(Item):
    """Base salary of a job offer."""

    #: Salary amount as it appears on the website.
    raw: Optional[str] = None

    #: The minimum value of the base salary as a number string.
    valueMin: Optional[str] = None

    #: The maximum value of the base salary as a number string.
    valueMax: Optional[str] = None

    #: The type of rate associated with the salary, e.g. monthly, annual, daily.
    rateType: Optional[str] = None

    #: Currency associated with the salary amount.
    currency: Optional[str] = None

    #: Currency associated with the salary amount, without normalization.
    currencyRaw: Optional[str] = None


@attrs.define(kw_only=True)
class HiringOrganization(Item):
    """Organization that is hiring for a job offer."""

    #: Name of the hiring organization.
    name: Optional[str] = None

    #: Organization information as available on the website.
    nameRaw: Optional[str] = None

    #: Identifier of the organization used by job posting website.
    id: Optional[str] = None



from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    AdditionalProperty,
    Address,
    AggregateRating,
    Amenity,
    ArticleListMetadata,
    ArticleMetadata,
    ArticleNavigationMetadata,
    Audio,
    Author,
    BaseSalary,
    Brand,
    Breadcrumb,
    BusinessPlaceMetadata,
    Gtin,
    HiringOrganization,
    Image,
    JobLocation,
    JobPostingMetadata,
    Link,
    NamedLink,
    OpeningHoursItem,
    ParentPlace,
    ProbabilityMetadata,
    ProbabilityRequest,
    ProductListMetadata,
    ProductMetadata,
    ProductNavigationMetadata,
    RealEstateArea,
    RealEstateMetadata,
    Request,
    StarRating,
    Video,
    cast_request,
)
from zyte_common_items.util import MetadataCaster, url_to_str


@attrs.define(slots=True, kw_only=True)
class ArticleFromList(Item):
    """Article from an article list from an article listing page.

    See :class:`ArticleList`.
    """

    #: Clean text of the article, including sub-headings, with newline
    #: separators.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the body
    #:   string),
    #: - line breaks included,
    #: - no length limit,
    #: - no normalization of Unicode characters.
    articleBody: Optional[str] = None

    #: All authors of the article.
    authors: Optional[List[Author]] = None

    #: Publication date of the article.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ" or
    #: "YYYY-MM-DDThh:mm:ss±zz:zz".
    #:
    #: With timezone, if available.
    #:
    #: If the actual publication date is not found, the date of the last
    #: modification is used instead.
    datePublished: Optional[str] = None

    #: Same date as
    #: :attr:`~zyte_common_items.ArticleFromList.datePublished`, but
    #: :before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: Headline or title.
    headline: Optional[str] = None

    #: Language of the article, as an ISO 639-1 language code.
    #:
    #: Sometimes the article language is not the same as the web page overall
    #: language.
    inLanguage: Optional[str] = None

    #: Main image.
    mainImage: Optional[Image] = None

    #: All images.
    images: Optional[List[Image]] = None

    #: Data extraction process metadata.
    metadata: Optional[ProbabilityMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ProbabilityMetadata)), kw_only=True  # type: ignore
    )

    #: Main URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(kw_only=True)
class Article(Item):
    #: Headline or title.
    headline: Optional[str] = None

    #: Publication date of the article.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ" or
    #: "YYYY-MM-DDThh:mm:ss±zz:zz".
    #:
    #: With timezone, if available.
    #:
    #: If the actual publication date is not found, the value of
    #: :attr:`~zyte_common_items.Article.dateModified` is used instead.
    datePublished: Optional[str] = None

    #: Same date as
    #: :attr:`~zyte_common_items.Article.datePublished`, but
    #: :before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: Date when the article was most recently modified.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ" or
    #: "YYYY-MM-DDThh:mm:ss±zz:zz".
    #:
    #: With timezone, if available.
    dateModified: Optional[str] = None

    #: Same date as
    #: :attr:`~zyte_common_items.Article.dateModified`, but
    #: :before parsing/normalization, i.e. as it appears on the website.
    dateModifiedRaw: Optional[str] = None

    #: All authors of the article.
    authors: Optional[List[Author]] = None

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Language of the article, as an ISO 639-1 language code.
    #:
    #: Sometimes the article language is not the same as the web page overall
    #: language.
    inLanguage: Optional[str] = None

    #: Main image.
    mainImage: Optional[Image] = None

    #: All images.
    images: Optional[List[Image]] = None

    #: A short summary of the article.
    #:
    #: It can be either human-provided (if available), or auto-generated.
    description: Optional[str] = None

    #: Clean text of the article, including sub-headings, with newline
    #: separators.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the body
    #:   string),
    #: - line breaks included,
    #: - no length limit,
    #: - no normalization of Unicode characters.
    articleBody: Optional[str] = None

    #: Simplified and standardized HTML of the article, including sub-headings,
    #: image captions and embedded content (videos, tweets, etc.).
    #:
    #: Format: HTML string normalized in a consistent way.
    articleBodyHtml: Optional[str] = None

    #: All videos.
    videos: Optional[List[Video]] = None

    #: All audios.
    audios: Optional[List[Audio]] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: The main URL of the article page.
    #:
    #: The URL of the final response, after any redirects.
    #:
    #: Required attribute.
    #:
    #: In case there is no article data on the page or the page was not
    #: reached, the returned "empty" item would still contain this URL field.
    url: str = attrs.field(converter=url_to_str)

    #: Data extraction process metadata.
    metadata: Optional[ArticleMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ArticleMetadata)), kw_only=True  # type: ignore
    )


@attrs.define(slots=True, kw_only=True)
class ArticleList(Item):
    """Article list from an article listing page.

    The :attr:`url` attribute is the only required attribute, all other fields
    are optional.
    """

    #: The main URL of the article list.
    #:
    #: The URL of the final response, after any redirects.
    #:
    #: Required attribute.
    #:
    #: In case there is no article list data on the page or the page was not
    #: reached, the returned item still contain this URL field and all the
    #: other available datapoints.
    url: str = attrs.field(converter=url_to_str)

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: List of article details found on the page.
    #:
    #: The order of the articles reflects their position on the page.
    articles: Optional[List[ArticleFromList]] = None

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Data extraction process metadata.
    metadata: Optional[ArticleListMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ArticleListMetadata)), kw_only=True  # type: ignore
    )


@attrs.define(kw_only=True)
class ProductVariant(Item):
    """:class:`Product` variant.

    See :attr:`Product.variants`.
    """

    #: List of name-value pais of data about a specific, otherwise unmapped
    #: feature.
    #:
    #: Additional properties usually appear in product pages in the form of a
    #: specification table or a free-form specification list.
    #:
    #: Additional properties that require 1 or more extra requests may not be
    #: extracted.
    #:
    #: See also ``features``.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: Availability status.
    #:
    #: The value is expected to be one of: ``"InStock"``, ``"OutOfStock"``.
    availability: Optional[str] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: Color.
    #:
    #: It is extracted as displayed (e.g. ``"white"``).
    #:
    #: See also ``size``, ``style``.
    color: Optional[str] = None

    #: Price currency `ISO 4217`_ alphabetic code (e.g. ``"USD"``).
    #:
    #: See also ``currencyRaw``.
    #:
    #: .. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217
    currency: Optional[str] = None

    #: Price currency as it appears on the webpage (no post-processing), e.g.
    #: ``"$"``.
    #:
    #: See also ``currency``.
    currencyRaw: Optional[str] = None

    #: List of standardized GTIN_ product identifiers associated with the
    #: product, which are unique for the product across different sellers.
    #:
    #: See also: ``mpn``, ``productId``, ``sku``.
    #:
    #: .. _GTIN: https://en.wikipedia.org/wiki/Global_Trade_Item_Number
    gtin: Optional[List[Gtin]] = None

    #: All product images.
    #:
    #: The main image (see ``mainImage``) should be first in the list.
    #:
    #: Images only displayed as part of the product description are excluded.
    images: Optional[List[Image]] = None

    #: Main product image.
    mainImage: Optional[Image] = None

    #: `Manufacturer part number (MPN)`_.
    #:
    #: A product should have the same MPN across different e-commerce websites.
    #:
    #: See also: ``gtin``, ``productId``, ``sku``.
    #:
    #: .. _Manufacturer part number (MPN): https://en.wikipedia.org/wiki/Part_number
    mpn: Optional[str] = None

    #: Name as it appears on the webpage (no post-processing).
    name: Optional[str] = None

    #: Price at which the product is being offered.
    #:
    #: It is a string with the price amount, with a full stop as decimal
    #: separator, and no thousands separator or currency (see ``currency`` and
    #: ``currencyRaw``), e.g. ``"10500.99"``.
    #:
    #: If ``regularPrice`` is not ``None``, ``price`` should always be lower
    #: than ``regularPrice``.
    price: Optional[str] = None

    #: Product identifier, unique within an e-commerce website.
    #:
    #: It may come in the form of an SKU or any other identifier, a hash, or
    #: even a URL.
    #:
    #: See also: ``gtin``, ``mpn``, ``sku``.
    productId: Optional[str] = None

    #: Price at which the product was being offered in the past, and which is
    #: presented as a reference next to the current price.
    #:
    #: It may be labeled as the original price, the list price, or the maximum
    #: retail price for which the product is sold.
    #:
    #: See ``price`` for format details.
    #:
    #: If ``regularPrice`` is not ``None``, it should always be higher than
    #: ``price``.
    regularPrice: Optional[str] = None

    #: Size or dimensions.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"XL"``).
    #:
    #: See also ``color``, ``style``.
    size: Optional[str] = None

    #: `Stock keeping unit (SKU)`_ identifier, i.e. a merchant-specific product
    #: identifier.
    #:
    #: See also: ``gtin``, ``mpn``, ``productId``.
    #:
    #: .. _Stock keeping unit (SKU): https://en.wikipedia.org/wiki/Stock_keeping_unit
    sku: Optional[str] = None

    #: Style.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"polka dots"``).
    #:
    #: See also ``color``, ``size``.
    style: Optional[str] = None

    #: Main URL from which the product variant data could be extracted.
    #:
    #: See also ``canonicalUrl``.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(kw_only=True)
class Product(Item):
    """Product from an e-commerce website.

    The :attr:`url` attribute is the only required attribute, all other fields
    are optional.
    """

    #: List of name-value pais of data about a specific, otherwise unmapped
    #: feature.
    #:
    #: Additional properties usually appear in product pages in the form of a
    #: specification table or a free-form specification list.
    #:
    #: Additional properties that require 1 or more extra requests may not be
    #: extracted.
    #:
    #: See also ``features``.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: Aggregate data about reviews and ratings.
    aggregateRating: Optional[AggregateRating] = None

    #: Availability status.
    #:
    #: The value is expected to be one of: ``"InStock"``, ``"OutOfStock"``.
    availability: Optional[str] = None

    #: Brand.
    brand: Optional[Brand] = None

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: Color.
    #:
    #: It is extracted as displayed (e.g. ``"white"``).
    #:
    #: See also ``size``, ``style``.
    color: Optional[str] = None

    #: Price currency `ISO 4217`_ alphabetic code (e.g. ``"USD"``).
    #:
    #: See also ``currencyRaw``.
    #:
    #: .. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217
    currency: Optional[str] = None

    #: Price currency as it appears on the webpage (no post-processing), e.g.
    #: ``"$"``.
    #:
    #: See also ``currency``.
    currencyRaw: Optional[str] = None

    #: Plain-text description.
    #:
    #: If the description is split across different parts of the source
    #: webpage, only the main part, containing the most useful pieces of
    #: information, should be extracted into this attribute.
    #:
    #: It may contain data found in other attributes (``features``,
    #: ``additionalProperties``).
    #:
    #: Format-wise:
    #:
    #: -   Line breaks and non-ASCII characters are allowed.
    #:
    #: -   There is no length limit for this attribute, the content should not
    #:     be truncated.
    #:
    #: -   There should be no whitespace at the beginning or end.
    #:
    #: See also ``descriptionHtml``.
    description: Optional[str] = None

    #: HTML description.
    #:
    #: See ``description`` for extraction details.
    #:
    #: The format is not the raw HTML from the source webpage. See the `HTML
    #: normalization specification`_ for details.
    #:
    #: .. _HTML normalization specification: https://docs.zyte.com/automatic-extraction/article.html#format-of-articlebodyhtml-field
    descriptionHtml: Optional[str] = None

    #: List of features.
    #:
    #: They are usually listed as bullet points in product webpages.
    #:
    #: See also ``additionalProperties``.
    features: Optional[List[str]] = None

    #: List of standardized GTIN_ product identifiers associated with the
    #: product, which are unique for the product across different sellers.
    #:
    #: See also: ``mpn``, ``productId``, ``sku``.
    #:
    #: .. _GTIN: https://en.wikipedia.org/wiki/Global_Trade_Item_Number
    gtin: Optional[List[Gtin]] = None

    #: All product images.
    #:
    #: The main image (see ``mainImage``) should be first in the list.
    #:
    #: Images only displayed as part of the product description are excluded.
    images: Optional[List[Image]] = None

    #: Main product image.
    mainImage: Optional[Image] = None

    #: Data extraction process metadata.
    metadata: Optional[ProductMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ProductMetadata)), kw_only=True  # type: ignore
    )

    #: `Manufacturer part number (MPN)`_.
    #:
    #: A product should have the same MPN across different e-commerce websites.
    #:
    #: See also: ``gtin``, ``productId``, ``sku``.
    #:
    #: .. _Manufacturer part number (MPN): https://en.wikipedia.org/wiki/Part_number
    mpn: Optional[str] = None

    #: Name as it appears on the webpage (no post-processing).
    name: Optional[str] = None

    #: Price at which the product is being offered.
    #:
    #: It is a string with the price amount, with a full stop as decimal
    #: separator, and no thousands separator or currency (see ``currency`` and
    #: ``currencyRaw``), e.g. ``"10500.99"``.
    #:
    #: If ``regularPrice`` is not ``None``, ``price`` should always be lower
    #: than ``regularPrice``.
    price: Optional[str] = None

    # Redefined to extend the documentation.
    #: Product identifier, unique within an e-commerce website.
    #:
    #: It may come in the form of an SKU or any other identifier, a hash, or
    #: even a URL.
    #:
    #: See also: ``gtin``, ``mpn``, ``sku``.
    productId: Optional[str] = None

    #: Price at which the product was being offered in the past, and which is
    #: presented as a reference next to the current price.
    #:
    #: It may be labeled as the original price, the list price, or the maximum
    #: retail price for which the product is sold.
    #:
    #: See ``price`` for format details.
    #:
    #: If ``regularPrice`` is not ``None``, it should always be higher than
    #: ``price``.
    regularPrice: Optional[str] = None

    #: Size or dimensions.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"XL"``).
    #:
    #: See also ``color``, ``style``.
    size: Optional[str] = None

    #: `Stock keeping unit (SKU)`_ identifier, i.e. a merchant-specific product
    #: identifier.
    #:
    #: See also: ``gtin``, ``mpn``, ``productId``.
    #:
    #: .. _Stock keeping unit (SKU): https://en.wikipedia.org/wiki/Stock_keeping_unit
    sku: Optional[str] = None

    #: Style.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"polka dots"``).
    #:
    #: See also ``color``, ``size``.
    style: Optional[str] = None

    #: Main URL from which the data has been extracted.
    #:
    #: See also ``canonicalUrl``.
    url: str = attrs.field(converter=url_to_str)

    #: List of variants.
    #:
    #: When slightly different versions of a product are displayed on the same
    #: product page, allowing you to choose a specific product version from a
    #: selection, each of those product versions are considered a product
    #: variant.
    #:
    #: Product variants usually differ in ``color`` or ``size``.
    #:
    #: The following items are *not* considered product variants:
    #:
    #: -   Different products within the same bundle of products.
    #:
    #: -   Product add-ons, e.g. premium upgrades of a base product.
    #:
    #: Only variant-specific data is extracted as product variant details. For
    #: example, if variant-specific versions of the product description do not
    #: exist in the source webpage, the description attributes of the product
    #: variant are *not* filled with the base product description.
    #:
    #: Extracted product variants may not include those that are not visible in
    #: the source webpage.
    #:
    #: Product variant details may not include those that require multiple
    #: additional requests (e.g. 1 or more requests per variant).
    variants: Optional[List[ProductVariant]] = None


@attrs.define(slots=True, kw_only=True)
class ProductFromList(Item):
    """Product from a product list from a product listing page of an e-commerce
    webpage.

    See :class:`ProductList`.
    """

    #: Price currency `ISO 4217`_ alphabetic code (e.g. ``"USD"``).
    #:
    #: See also ``currencyRaw``.
    #:
    #: .. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217
    currency: Optional[str] = None

    #: Price currency as it appears on the webpage (no post-processing), e.g.
    #: ``"$"``.
    #:
    #: See also ``currency``.
    currencyRaw: Optional[str] = None

    #: Main product image.
    mainImage: Optional[Image] = None

    #: Data extraction process metadata.
    metadata: Optional[ProbabilityMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ProbabilityMetadata)), kw_only=True  # type: ignore
    )

    #: Name as it appears on the webpage (no post-processing).
    name: Optional[str] = None

    #: Price at which the product is being offered.
    #:
    #: It is a string with the price amount, with a full stop as decimal
    #: separator, and no thousands separator or currency (see ``currency`` and
    #: ``currencyRaw``), e.g. ``"10500.99"``.
    #:
    #: If ``regularPrice`` is not ``None``, ``price`` should always be lower
    #: than ``regularPrice``.
    price: Optional[str] = None

    #: Product identifier, unique within an e-commerce website.
    #:
    #: It may come in the form of an SKU or any other identifier, a hash, or
    #: even a URL.
    productId: Optional[str] = None

    #: Price at which the product was being offered in the past, and which is
    #: presented as a reference next to the current price.
    #:
    #: It may be labeled as the original price, the list price, or the maximum
    #: retail price for which the product is sold.
    #:
    #: See ``price`` for format details.
    #:
    #: If ``regularPrice`` is not ``None``, it should always be higher than
    #: ``price``.
    regularPrice: Optional[str] = None

    #: Main URL from which the product data could be extracted.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(slots=True, kw_only=True)
class ProductList(Item):
    """Product list from a product listing page of an e-commerce webpage.

    It represents, for example, a single page from a category.

    The :attr:`url` attribute is the only required attribute, all other fields
    are optional.
    """

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: Name of the product listing as it appears on the webpage
    #: (no post-processing).
    #:
    #: For example, if the webpage is one of the pages of the Robots category,
    #: ``categoryName`` is ``'Robots'``.
    categoryName: Optional[str] = None

    #: Data extraction process metadata.
    metadata: Optional[ProductListMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ProductListMetadata)), kw_only=True  # type: ignore
    )

    #: Current page number, if displayed explicitly on the list page.
    #:
    #: Numeration starts with 1.
    pageNumber: Optional[int] = None

    #: Link to the next page.
    paginationNext: Optional[Link] = None

    #: List of products.
    #:
    #: It only includes product information found in the product listing page
    #: itself. Product information that requires visiting each product URL is
    #: not meant to be covered.
    #:
    #: The order of the products reflects their position on the rendered page.
    #: Product order is top-to-bottom, and left-to-right or right-to-left
    #: depending on the webpage locale.
    products: Optional[List[ProductFromList]] = None

    #: Main URL from which the data has been extracted.
    #:
    #: See also ``canonicalUrl``.
    url: str = attrs.field(converter=url_to_str)


@attrs.define(slots=True, kw_only=True)
class BusinessPlace(Item):
    """Business place, with properties typically seen on maps or business
    listings."""

    #: Unique identifier of the place on the website.
    placeId: Optional[str] = None

    #: The main URL that the place data was extracted from.
    #:
    #: The URL of the final response, after any redirects.
    #:
    #: In case there is no product data on the page or the page was not reached, the returned "empty"
    #: item would still contain url field and metadata field with dateDownloaded.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: The name of the place.
    name: Optional[str] = None

    #: List of actions that can be performed directly from the URLs on the place page, including URLs.
    actions: Optional[List[NamedLink]] = None

    #: List of name-value pais of any unmapped additional properties specific to the place.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: The address details of the place.
    address: Optional[Address] = None

    #: The details of the reservation action,
    #: e.g. table reservation in case of restaurants
    #: or room reservation in case of hotels.
    reservationAction: Optional[NamedLink] = None

    #: List of categories the place belongs to.
    categories: Optional[List[str]] = None

    #: The description of the place.
    #:
    #: Stripped of white spaces.
    description: Optional[str] = None

    #: List of frequently mentioned features of this place.
    features: Optional[List[str]] = None

    #: URL to a map of the place.
    map: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #: A list of URL values of all images of the place.
    images: Optional[List[Image]] = None

    #: List of amenities of the place.
    amenityFeatures: Optional[List[Amenity]] = None

    #: The overall rating, based on a collection of reviews or ratings.
    aggregateRating: Optional[AggregateRating] = None

    #: Official star rating of the place.
    starRating: Optional[StarRating] = None

    #: If the place is located inside another place, these are the details of the parent place.
    containedInPlace: Optional[ParentPlace] = None

    #: Ordered specification of opening hours, including data for opening and closing time for each day of the week.
    openingHours: Optional[List[OpeningHoursItem]] = None

    #: List of partner review sites.
    reviewSites: Optional[List[NamedLink]] = None

    #: The phone number associated with the place, as it appears on the page.
    telephone: Optional[str] = None

    #: How is the price range of the place viewed by its customers (from z to zzzz).
    priceRange: Optional[str] = None

    #: Which timezone is the place situated in.
    #:
    #: Standard: Name compliant with IANA tz database (tzdata).
    timezone: Optional[str] = None

    #: If the information is verified by the owner of this place.
    isVerified: Optional[bool] = None

    #: The URL pointing to the official website of the place.
    website: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )

    #:  List of the tags associated with the place.
    tags: Optional[List[str]] = None

    #: Data extraction process metadata.
    metadata: Optional[BusinessPlaceMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(BusinessPlaceMetadata)), kw_only=True  # type: ignore
    )


@attrs.define(slots=True, kw_only=True)
class RealEstate(Item):
    #: The url of the final response, after any redirects.
    url: str = attrs.field(converter=url_to_str)

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: The identifier of the real estate, usually assigned by the seller and unique within a website, similar to product SKU.
    realEstateId: Optional[str] = None

    #: The name of the real estate.
    name: Optional[str] = None

    #: Publication date of the real estate offer.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    datePublished: Optional[str] = None

    #: Same date as datePublished, but before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: The description of the real estate.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string),
    #:
    #: - line breaks included,
    #:
    #: - no length limit,
    #:
    #: - no normalization of Unicode characters,
    #:
    #: - no concatenation of description from different parts of the page.
    description: Optional[str] = None

    #: The details of the main image of the real estate.
    mainImage: Optional[Image] = None

    #: A list of URL values of all images of the real estate.
    images: Optional[List[Image]] = None

    #: The details of the address of the real estate.
    address: Optional[Address] = None

    #: Real estate area details.
    area: Optional[RealEstateArea] = None

    #: The total number of bathrooms in the real estate.
    numberOfBathroomsTotal: Optional[int] = None

    #: The number of full bathrooms in the real estate.
    numberOfFullBathrooms: Optional[int] = None

    #: The number of partial bathrooms in the real estate.
    numberOfPartialBathrooms: Optional[int] = None

    #: The number of bedrooms in the real estate.
    numberOfBedrooms: Optional[int] = None

    #: The number of rooms (excluding bathrooms and closets) of the real estate.
    numberOfRooms: Optional[int] = None

    #: Type of a trade action: buying or renting.
    tradeType: Optional[str] = None

    #: The offer price of the real estate.
    price: Optional[str] = None

    #: The rental period to which the rental price applies, only available in case of rental. Usually weekly, monthly, quarterly, yearly.
    rentalPeriod: Optional[str] = None

    #: Currency associated with the price, as appears on the page (no post-processing).
    currencyRaw: Optional[str] = None

    #: The currency of the price, in 3-letter ISO 4217 format.
    currency: Optional[str] = None

    #: A name-value pair field holding information pertaining to specific features. Usually in a form of a specification table or freeform specification list.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: Type of the property, e.g. flat, house, land.
    propertyType: Optional[str] = None

    #: The year the real estate was built.
    yearBuilt: Optional[int] = None

    #: The URL of the virtual tour of the real estate.
    virtualTourUrl: Optional[str] = None

    #: Contains metadata about the data extraction process.
    metadata: Optional[RealEstateMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(RealEstateMetadata)), kw_only=True  # type: ignore
    )


class RequestListCaster:
    def __init__(self, target):
        self._target = target

    def __call__(self, value):
        return [cast_request(item, self._target) for item in value]


@attrs.define(kw_only=True)
class ProductNavigation(Item):
    """Represents the navigational aspects of a product listing page on an
    e-commerce website"""

    #: Main URL from which the data is extracted.
    url: str = attrs.field(converter=url_to_str)

    #: Name of the category/page with the product list.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string)
    categoryName: Optional[str] = None

    #: List of sub-category links ordered by their position in the page.
    subCategories: Optional[List[ProbabilityRequest]] = attrs.field(
        default=None, converter=attrs.converters.optional(RequestListCaster(ProbabilityRequest)), kw_only=True  # type: ignore
    )

    #: List of product links found on the page category ordered by their position in the page.
    items: Optional[List[ProbabilityRequest]] = attrs.field(
        default=None, converter=attrs.converters.optional(RequestListCaster(ProbabilityRequest)), kw_only=True  # type: ignore
    )

    #: A link to the next page, if available.
    nextPage: Optional[Request] = None

    #: Number of the current page.
    #:
    #: It should only be extracted if the webpage shows a page number.
    #:
    #: It must be 1-based. For example, if the first page of a listing is
    #: numbered as 0 on the website, it should be extracted as `1` nonetheless.
    pageNumber: Optional[int] = None

    #: Data extraction process metadata.
    metadata: Optional[ProductNavigationMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ProductNavigationMetadata)), kw_only=True  # type: ignore
    )


@attrs.define(kw_only=True)
class ArticleNavigation(Item):
    """Represents the navigational aspects of an article listing webpage.

    See :class:`ArticleList`.
    """

    #: Main URL from which the data is extracted.
    url: str = attrs.field(converter=url_to_str)

    #: Name of the category/page.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string)
    categoryName: Optional[str] = None

    #: List of sub-category links ordered by their position in the page.
    subCategories: Optional[List[ProbabilityRequest]] = attrs.field(
        default=None, converter=attrs.converters.optional(RequestListCaster(ProbabilityRequest)), kw_only=True  # type: ignore
    )

    #: Links to listed items in order of appearance.
    items: Optional[List[ProbabilityRequest]] = attrs.field(
        default=None, converter=attrs.converters.optional(RequestListCaster(ProbabilityRequest)), kw_only=True  # type: ignore
    )

    #: A link to the next page, if available.
    nextPage: Optional[Request] = None

    #: Number of the current page.
    #:
    #: It should only be extracted if the webpage shows a page number.
    #:
    #: It must be 1-based. For example, if the first page of a listing is
    #: numbered as 0 on the website, it should be extracted as `1` nonetheless.
    pageNumber: Optional[int] = None

    #: Data extraction process metadata.
    metadata: Optional[ArticleNavigationMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(ArticleNavigationMetadata)), kw_only=True  # type: ignore
    )


@attrs.define(kw_only=True)
class JobPosting(Item):
    #: The url of the final response, after any redirects.
    url: str = attrs.field(converter=url_to_str)

    #: The identifier of the job posting.
    jobPostingId: Optional[str] = None

    #: Publication date of the job posting.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    datePublished: Optional[str] = None

    #: Same date as datePublished, but before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: The date when the job posting was most recently modified.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    dateModified: Optional[str] = None

    #: Same date as dateModified, but before parsing/normalization, i.e. as it appears on the website.
    dateModifiedRaw: Optional[str] = None

    #: The date after which the job posting is not valid, e.g. the end of an offer.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    validThrough: Optional[str] = None

    #: Same date as validThrough, but before parsing/normalization, i.e. as it appears on the website.
    validThroughRaw: Optional[str] = None

    #: The title of the job posting.
    jobTitle: Optional[str] = None

    #: The headline of the job posting.
    headline: Optional[str] = None

    #: A (typically single) geographic location associated with the job position.
    jobLocation: Optional[JobLocation] = None

    #: A description of the job posting including sub-headings, with newline separators.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string),
    #:
    #: - line breaks included,
    #:
    #: - no length limit,
    #:
    #: - no normalization of Unicode characters.
    description: Optional[str] = None

    #: Simplified HTML of the description, including sub-headings, image captions and embedded content.
    descriptionHtml: Optional[str] = None

    #: Type of employment (e.g. full-time, part-time, contract, temporary, seasonal, internship).
    employmentType: Optional[str] = None

    #: The base salary of the job or of an employee in the proposed role.
    baseSalary: Optional[BaseSalary] = None

    #: Candidate requirements for the job.
    requirements: Optional[List[str]] = None

    #: Information about the organization offering the job position.
    hiringOrganization: Optional[HiringOrganization] = None

    #: Job start date
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    jobStartDate: Optional[str] = None

    #: Same date as jobStartDate, but before parsing/normalization, i.e. as it appears on the website.
    jobStartDateRaw: Optional[str] = None

    #: Specifies the remote status of the position.
    remoteStatus: Optional[str] = None

    #: Contains metadata about the data extraction process.
    metadata: Optional[JobPostingMetadata] = attrs.field(
        default=None, converter=attrs.converters.optional(MetadataCaster(JobPostingMetadata)), kw_only=True  # type: ignore
    )



import pytest
from lxml.html import fromstring
from parsel import Selector, SelectorList
from price_parser import Price
from web_poet import HttpResponse, field
from zyte_parsers import Breadcrumb as zp_Breadcrumb
from zyte_parsers import extract_breadcrumbs

from zyte_common_items import BasePage, Breadcrumb, ProductPage
from zyte_common_items.processors import (
    _format_price,
    brand_processor,
    breadcrumbs_processor,
)

breadcrumbs_html = """
<div class="pagesbar">
  <ul>
    <li><a href="/blog/">Home</a></li>
    <li><a href="/blog/about/">About</a></li>
  </ul>
</div>
"""

breadcrumbs_expected = [
    Breadcrumb(name="Home", url="http://www.example.com/blog/"),
    Breadcrumb(name="About", url="http://www.example.com/blog/about/"),
]


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ([], []),
        ("foo", "foo"),
        (Selector(text="<html></html>"), None),
        (SelectorList([]), None),
        (breadcrumbs_expected, breadcrumbs_expected),
        (fromstring(breadcrumbs_html), breadcrumbs_expected),
        (Selector(text=breadcrumbs_html), breadcrumbs_expected),
        (
            [
                zp_Breadcrumb(name="Home", url="http://www.example.com/blog/"),
                zp_Breadcrumb(name="About", url="http://www.example.com/blog/about/"),
            ],
            breadcrumbs_expected,
        ),
    ],
)
def test_breadcrumbs(input_value, expected_value):
    base_url = "http://www.example.com/blog/"

    class BreadcrumbsPage(BasePage):
        @field(out=[breadcrumbs_processor])
        def breadcrumbs(self):
            return input_value

    page = BreadcrumbsPage(base_url)  # type: ignore[arg-type]
    assert page.breadcrumbs == expected_value


def test_breadcrumbs_page():
    class MyProductPage(ProductPage):
        @field
        def breadcrumbs(self):
            return self.css(".pagesbar")

    response = HttpResponse(
        url="http://www.example.com/blog/",
        body=f"<html><body>{breadcrumbs_html}</body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.breadcrumbs == breadcrumbs_expected


def test_breadcrumbs_page_zyte_parsers():
    class MyProductPage(ProductPage):
        @field
        def breadcrumbs(self):
            return extract_breadcrumbs(self.css(".pagesbar")[0], base_url=self.url)

    response = HttpResponse(
        url="http://www.example.com/blog/",
        body=f"<html><body>{breadcrumbs_html}</body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.breadcrumbs == breadcrumbs_expected


def test_breadcrumbs_base_url():
    class MyProductPage(ProductPage):
        @field
        def breadcrumbs(self):
            return self.css(".pagesbar")

    breadcrumbs_html_base = """
    <html>
    <head>
        <base href="http://www.example.com/blog/">
    </head>
    <body>
    <div class="pagesbar">
      <ul>
        <li><a href=".">Home</a></li>
        <li><a href="about/">About</a></li>
      </ul>
    </div>
    </body>
    </html>
    """

    response = HttpResponse(
        url="http://www.example.com/",
        body=breadcrumbs_html_base.encode(),
    )
    page = MyProductPage(response=response)
    assert page.breadcrumbs == breadcrumbs_expected


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ("", ""),
        ("foo", "foo"),
        (Selector(text="<html></html>"), None),
        (SelectorList([]), None),
        (fromstring("<p>foo</p>"), "foo"),
        (fromstring("<img alt='foo'>"), "foo"),
        (fromstring("<p><img alt='foo'></p>"), "foo"),
        (fromstring("<p><p><img alt='foo'></p></p>"), "foo"),
        (Selector(text="<p>foo</p>"), "foo"),
        (SelectorList([Selector(text="<p>foo</p>")]), "foo"),
    ],
)
def test_brand(input_value, expected_value):
    class BrandPage(BasePage):
        @field(out=[brand_processor])
        def brand(self):
            return input_value

    page = BrandPage("http://www.example.com/")  # type: ignore[arg-type]
    assert page.brand == expected_value


def test_brand_page():
    class MyProductPage(ProductPage):
        @field
        def brand(self):
            return self.css("body")

    response = HttpResponse(
        url="http://www.example.com/",
        body="<html><body><img alt='foo'></body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.brand == "foo"


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        ("", None),
        ("NaN", None),
        ("Infinity", None),
        ("1", "1.00"),
        ("1.2", "1.20"),
        ("1.23", "1.23"),
        # Fails due to . being considered a thousands separator in this case.
        # ("1.234", "1.234"),
        ("1.2345", "1.2345"),
    ],
)
def test_format_price(input_value, expected_value):
    assert _format_price(Price.fromstring(input_value)) == expected_value



"""This test file is ignored by mypy so don't rely on type assertions in this.

Instead, take a look at the ``tests/test_mypy.py``.
"""
import pytest
from web_poet import HttpResponse, RequestUrl, ResponseUrl

from zyte_common_items import (
    Breadcrumb,
    Image,
    Link,
    Product,
    ProductFromList,
    ProductList,
    ProductVariant,
)


@pytest.mark.parametrize(
    "cls,fields",
    [
        (Image, ["url"]),
        (Breadcrumb, ["url"]),
        (Link, ["url"]),
        (Product, ["url", "canonicalUrl"]),
        (ProductList, ["url", "canonicalUrl"]),
        (ProductFromList, ["url"]),
        (ProductVariant, ["url", "canonicalUrl"]),
    ],
)
def test_webpoet_URL_classes(cls, fields):
    """Ensure that the URL classes from web-poet are properly converted into
    URL strings when instantiating.
    """
    url = "https://www.some-url.com"
    response = HttpResponse(url, b"")

    response_url_obj = response.url
    assert type(response_url_obj) == ResponseUrl

    request_url_obj = response.urljoin("/another-page")
    assert type(request_url_obj) == RequestUrl

    # Ensure that both types of URL classes are covered
    for url_obj in [response_url_obj, request_url_obj]:
        data = {field: url_obj for field in fields}
        obj = cls(**data)

        for field in fields:
            # The URL classes should be converted to strings on instantiation
            attribute = getattr(obj, field)
            assert type(attribute) == str
            assert attribute == str(url_obj)

            # The conversion should also work via field assignment
            setattr(obj, field, response.urljoin("/somewhere"))
            attribute = getattr(obj, field)
            assert type(attribute) == str
            assert attribute == "https://www.some-url.com/somewhere"

            # Setting other values that are not strings or URL classes would
            # raise a ValueError
            with pytest.raises(ValueError):
                setattr(obj, field, 123)


@pytest.mark.parametrize("cls", [ProductVariant, Product, ProductFromList])
def test_webpoet_URL_mainImage(cls):
    response = HttpResponse("https://www.some-page", b"")
    data = {
        "mainImage": {
            "url": response.urljoin("img1.png"),
        },
        "url": response.url,
    }
    assert type(data["mainImage"]["url"]) == RequestUrl

    # The URL classes should be converted to strings on instantiation
    obj = cls.from_dict(data)
    assert type(obj.mainImage) == Image
    assert obj.mainImage.url == "https://www.some-page/img1.png"

    # The conversion should also work via field assignment
    img_url = response.urljoin("/different-img.png")
    assert type(img_url) == RequestUrl
    obj.mainImage.url = img_url
    assert type(obj.mainImage) == Image
    assert obj.mainImage.url == "https://www.some-page/different-img.png"

    # Setting other values that are not strings or URL classes would
    # raise a ValueError
    with pytest.raises(ValueError):
        obj.mainImage.url = False

    data = {"mainImage": {"url": 123}, "url": 123}
    with pytest.raises(ValueError):
        obj = cls.from_dict(data)


@pytest.mark.parametrize("cls", [ProductVariant, Product])
def test_webpoet_URL_images(cls):
    response = HttpResponse("https://www.some-page", b"")
    data = {
        "images": [
            {"url": response.urljoin("img1.png")},
            {"url": response.urljoin("img2.png")},
        ],
        "url": response.url,
    }
    assert type(data["images"][0]["url"]) == RequestUrl

    # The URL classes should be converted to strings on instantiation
    obj = cls.from_dict(data)
    assert type(obj.images[0]) == Image
    assert [img.url for img in obj.images] == [
        "https://www.some-page/img1.png",
        "https://www.some-page/img2.png",
    ]

    # The conversion should also work via field assignment
    img_url = response.urljoin("/different-img.png")
    assert type(img_url) == RequestUrl
    obj.images[0].url = img_url
    assert type(obj.images[0]) == Image
    assert obj.images[0].url == "https://www.some-page/different-img.png"

    # Setting other values that are not strings or URL classes would
    # raise a ValueError
    with pytest.raises(ValueError):
        obj.images[1].url = False

    data = {
        "images": [
            {"url": 123},
            {"url": 456},
        ],
        "url": 789,
    }
    with pytest.raises(ValueError):
        obj = cls.from_dict(data)



from typing import List, Optional, Union

import attrs
import pytest

from zyte_common_items import Item, Product, is_data_container


class NotConsideredAnItem:
    """It has to inherit from Item to be considered one."""

    pass


@attrs.define
class SubItem(Item):
    name: str


@attrs.define
class BigItem(Item):
    sub_item: Optional[SubItem] = None


def test_is_data_container():
    """It should be able to discern if a given class/instance is a data container
    that is defined inside this repository.
    """
    assert is_data_container(Item)
    assert is_data_container(SubItem)
    assert is_data_container(Item())
    assert is_data_container(SubItem(name="test"))

    assert not is_data_container(NotConsideredAnItem)
    assert not is_data_container(NotConsideredAnItem())


def test_from_dict():
    """Tests the annotation-aware behavior for correcty deriving the correct type
    in the sub fields.
    """

    assert BigItem(sub_item=None).sub_item is None

    item = BigItem.from_dict({"sub_item": {"name": "hello"}})
    type(item.sub_item) == SubItem  # noqa: B015


def test_from_dict_bad_annotation():
    """Items with fields annotated with a Union of multiple different types should
    error out.
    """

    @attrs.define
    class A(Item):
        a: Union[int, str]

    pattern = (
        r"^tests\.\S+\.A\.a is annotated with typing\.Union\[int, str\]\. "
        r"Fields should only be annotated with one type \(or optional\)\.$"
    )
    with pytest.raises(ValueError, match=pattern):
        A.from_dict({"a": 1})


def test_from_dict_non_dict():
    @attrs.define
    class A(Item):
        a: str

    pattern = r"Expected a dict with fields from tests\.\S+?\.A, got 'a'\."
    with pytest.raises(ValueError, match=pattern):
        A.from_dict("a")  # type: ignore


def test_from_dict_non_dict_field():
    @attrs.define
    class B(Item):
        b: str

    @attrs.define
    class A(Item):
        a: B

    pattern = r"Expected a to be a dict with fields from tests\.\S+?\.B, got 'b'\."
    with pytest.raises(ValueError, match=pattern):
        A.from_dict({"a": "b"})


def test_from_dict_from_list_non_list_field():
    @attrs.define
    class A(Item):
        a: List[str]

    pattern = r"Expected a to be a list, got 'b'\."
    with pytest.raises(ValueError, match=pattern):
        A.from_dict({"a": "b"})


def test_from_dict_from_list_non_dict_field():
    @attrs.define
    class B(Item):
        b: str

    @attrs.define
    class A(Item):
        a: List[B]

    pattern = r"Expected a\[0\] to be a dict with fields from tests\.\S+?\.B, got 'b'\."
    with pytest.raises(ValueError, match=pattern):
        A.from_dict({"a": ["b"]})


def test_item_unknown_input():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating=dict(worstRating=0),
            url="https://example.com/?product=product22",
        )
    )
    assert product._unknown_fields_dict["a"] == "b"
    assert product.aggregateRating._unknown_fields_dict["worstRating"] == 0
    assert product.additionalProperties[0]._unknown_fields_dict["max"] == 10


def test_item_attribute_error():
    foo = SubItem(name="foo")
    with pytest.raises(AttributeError):
        foo.value = "bar"  # type: ignore[attr-defined]


def test_item_unknown_field_init():
    with pytest.raises(TypeError):
        SubItem(name="foo", value="bar")  # type: ignore[call-arg]



import attrs
import pytest

from zyte_common_items.util import split_in_unknown_and_known_fields


@attrs.define(auto_attribs=True, slots=True)
class _TestItem:
    k1: int
    k2: int


def test_split_in_unknown_and_known_fields():
    input = dict(k1=1, k2=2, extra=3)
    unknown, known = split_in_unknown_and_known_fields(input, _TestItem)
    item = _TestItem(**known)
    assert attrs.asdict(item) == dict(k1=1, k2=2)
    assert unknown == dict(extra=3)

    unknown, known = split_in_unknown_and_known_fields(known, _TestItem)
    assert unknown == {}

    for empty_input in ({}, None):  # type: ignore[var-annotated]
        ret = split_in_unknown_and_known_fields(empty_input, _TestItem)
        assert ret == ({}, {})

    with pytest.raises(ValueError):
        split_in_unknown_and_known_fields(input, str)



import pytest
from web_poet import HttpResponse, field

from zyte_common_items import (
    BusinessPlacePage,
    JobPostingPage,
    ProductPage,
    RealEstatePage,
)

HTML = b"""
<!DOCTYPE html>
<html>
    <body>
        <article><div>\xe2\x82\xact&eacute;xt</div><p style="color:blue">text2&gt;</p><div class="twitter-tweet">tw</div></article>
    </body>
</html>
"""


DESCR_HTML_CLEANED = """<article>

<p>€téxt</p>

<p>text2&gt;</p>

<div class="twitter-tweet">tw</div>

</article>"""


TEXT_CLEANED = "€téxt\n\ntext2>\n\ntw"


DESCR_HTML_WRAPPED = """<article>

<p>€téxt</p>

<p>text2&gt;</p>

<p>tw</p>

</article>"""


@pytest.mark.asyncio
async def test_descriptionHtml_selector():
    # descriptionHtml is extracted from the returned node
    # description is produced from descriptionHtml (with html-text)
    class CustomProductPage(ProductPage):
        call_count = 0

        @field
        def descriptionHtml(self):
            self.call_count += 1
            return self.css("article")

    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert page.descriptionHtml == DESCR_HTML_CLEANED
    assert page.call_count == 1
    assert await page.description == TEXT_CLEANED
    assert page.call_count == 1

    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert await page.description == TEXT_CLEANED
    assert page.call_count == 1
    assert page.descriptionHtml == DESCR_HTML_CLEANED
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_descriptionHtml_explicit():
    # descriptionHtml is returned as is
    # description is produced from descriptionHtml
    class CustomProductPage(ProductPage):
        call_count = 0

        @field
        def descriptionHtml(self):
            self.call_count += 1
            return self.css("article").get()

    html = '<article><div>€téxt</div><p style="color:blue">text2&gt;</p><div class="twitter-tweet">tw</div></article>'

    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert page.descriptionHtml == html
    assert page.call_count == 1
    assert await page.description == TEXT_CLEANED
    assert page.call_count == 2  # we want this to be 1
    assert await page.description == TEXT_CLEANED
    assert page.call_count == 2  # we want this to be 1

    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert await page.description == TEXT_CLEANED
    assert page.call_count == 1
    assert page.descriptionHtml == html
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_descriptionHtml_async():
    class CustomProductPage(ProductPage):
        @field
        async def descriptionHtml(self):
            return self.css("article")

    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert await page.descriptionHtml == DESCR_HTML_CLEANED
    assert await page.description == TEXT_CLEANED

    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert await page.description == TEXT_CLEANED
    assert await page.descriptionHtml == DESCR_HTML_CLEANED


@pytest.mark.asyncio
async def test_description_selector():
    # description is extracted from the returned node
    # descriptionHtml is extracted from the same node
    class CustomProductPage(ProductPage):
        call_count = 0

        @field
        def description(self):
            self.call_count += 1
            return self.css("article")

    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert page.description == TEXT_CLEANED
    assert page.call_count == 1
    assert await page.descriptionHtml == DESCR_HTML_CLEANED
    assert page.call_count == 1

    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert await page.descriptionHtml == DESCR_HTML_CLEANED
    assert page.call_count == 1
    assert page.description == TEXT_CLEANED
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_description_explicit():
    # description is returned as is
    # descriptionHtml is produced by wrapping description
    class CustomProductPage(ProductPage):
        call_count = 0

        @field
        def description(self):
            self.call_count += 1
            return TEXT_CLEANED

    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert page.description == TEXT_CLEANED
    assert page.call_count == 1
    assert await page.descriptionHtml == DESCR_HTML_WRAPPED
    assert page.call_count == 1
    assert await page.descriptionHtml == DESCR_HTML_WRAPPED
    assert page.call_count == 1

    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert page.call_count == 0
    assert await page.descriptionHtml == DESCR_HTML_WRAPPED
    assert page.call_count == 1
    assert page.description == TEXT_CLEANED
    assert page.call_count == 2  # we want this to be 1


@pytest.mark.asyncio
async def test_none():
    # when neither description nor descriptionHtml are user-defined,
    # both return None
    class CustomProductPage(ProductPage):
        pass

    url = "https://example.com"
    page = CustomProductPage(response=HttpResponse(url=url, body=HTML))
    assert await page.description is None
    assert await page.descriptionHtml is None


@pytest.mark.asyncio
async def test_job_posting_mixin():
    class CustomPage(JobPostingPage):
        @field
        def descriptionHtml(self):
            return self.css("article")

    url = "https://example.com"
    page = CustomPage(response=HttpResponse(url=url, body=HTML))
    assert page.descriptionHtml == DESCR_HTML_CLEANED
    assert await page.description == TEXT_CLEANED


@pytest.mark.parametrize(
    "page_class",
    (
        BusinessPlacePage,
        JobPostingPage,
        RealEstatePage,
    ),
)
def test_description_simple(page_class: type):
    class CustomPage(page_class):
        @field
        def description(self):
            return self.css("article")

    url = "https://example.com"
    page = CustomPage(response=HttpResponse(url=url, body=HTML))
    assert page.description == TEXT_CLEANED



from collections import deque
from collections.abc import Collection
from contextlib import contextmanager

# In Python ≤ 3.8 you cannot annotate with “collections.abc.Collection[Item]”,
# so we need to import typing.Collection for annotation instead.
from typing import Collection as CollectionType
from typing import Deque, Optional, Type, cast

import attrs
import pytest
from itemadapter import ItemAdapter
from itemadapter.adapter import AdapterInterface

from zyte_common_items import Item, Product, ZyteItemAdapter
from zyte_common_items.adapter import ZyteItemKeepEmptyAdapter

from .test_items import _PRODUCT_ALL_KWARGS, _PRODUCT_MIN_KWARGS


@contextmanager
def configured_adapter(adapter=ZyteItemAdapter):
    ItemAdapter.ADAPTER_CLASSES.appendleft(adapter)
    try:
        yield
    finally:
        ItemAdapter.ADAPTER_CLASSES.popleft()


def test_asdict_all_fields():
    product = Product(**_PRODUCT_ALL_KWARGS)
    with configured_adapter():
        adapter = ItemAdapter(product)
        actual_dict = adapter.asdict()
    expected_dict = {
        "additionalProperties": [dict(name="foo", value="bar")],
        "aggregateRating": dict(
            bestRating=5.0,
            ratingValue=2.5,
            reviewCount=123,
        ),
        "availability": "InStock",
        "brand": dict(name="Ka-pow"),
        "breadcrumbs": [
            dict(name="Level 1", url="http://example.com/level1"),
            dict(name="Level 2", url="http://example.com/level1/level2"),
        ],
        "canonicalUrl": "https://example.com/product22",
        "color": "white",
        "currency": "USD",
        "currencyRaw": "$",
        "description": "Full freshness all over the fridge\n5 Conversion Modes on demand\nSuper Cooling Plus™",
        "descriptionHtml": (
            "<article><p>Full freshness all over the fridge</p>"
            "<p>5 Conversion Modes on demand</p>"
            "<p>Super Cooling Plus&trade;</p></article>"
        ),
        "features": ["Easily store fragile products.", "Bluetooth connectivity."],
        "gtin": [dict(type="foo", value="bar")],
        "images": [
            dict(url="http://example.com/image1.png"),
        ],
        "mainImage": dict(url="http://example.com/image1.png"),
        "metadata": dict(
            dateDownloaded="2022-12-31T13:01:54Z",
            probability=1.0,
        ),
        "mpn": "HSC0424PP",
        "name": "White two-door refrigerator",
        "price": "9999.99",
        "productId": "A123DK9823",
        "regularPrice": "11999.99",
        "size": "XL",
        "sku": "A123DK9823",
        "style": "polka dots",
        "url": "https://example.com/?product=product22",
        "variants": [{}],
    }
    assert actual_dict == expected_dict


def test_product_asdict_min():
    product = Product(**_PRODUCT_MIN_KWARGS)
    with configured_adapter():
        adapter = ItemAdapter(product)
        actual_dict = adapter.asdict()
    expected_dict = {
        "url": "https://example.com/?product=product22",
    }
    assert actual_dict == expected_dict


def test_product_asdict_list():
    product = Product(**_PRODUCT_MIN_KWARGS, variants=[{}])  # type: ignore[list-item]
    with configured_adapter():
        adapter = ItemAdapter(product)
        actual_dict = adapter.asdict()
    expected_dict = {
        "url": "https://example.com/?product=product22",
        "variants": [{}],
    }
    assert actual_dict == expected_dict


class EmptyCollection(Collection):
    def __contains__(self, item):
        return False

    def __len__(self):
        return 0

    def __iter__(self):
        return
        yield


@pytest.mark.parametrize(
    "value",
    (
        [],
        tuple(),
        EmptyCollection(),
    ),
)
def test_asdict_empty_collection(value):
    @attrs.define
    class _Item(Item):
        children: CollectionType[Item]

    item = _Item(value)
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual_dict = adapter.asdict()
    assert actual_dict == {}


def test_asdict_unknown_fields():
    input_dict = dict(
        a="b",
        additionalProperties=[{"name": "a", "value": "b", "max": 10}],
        aggregateRating={"worstRating": 0},
        url="https://example.com/",
    )
    product = Product.from_dict(input_dict)
    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter.asdict() == input_dict


def test_field_meta():
    metadata = {"b": "c"}

    @attrs.define
    class _Item(Item):
        a = attrs.field(default=None, metadata=metadata)

    item = _Item()
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual_dict = adapter.get_field_meta("a")
    assert actual_dict == metadata


def test_field_meta_missing_field():
    item = Item()
    with configured_adapter():
        adapter = ItemAdapter(item)
        with pytest.raises(KeyError):
            adapter.get_field_meta("a")


def test_field_meta_unknown_field():
    """There cannot be metadata for unknown fields, but trying to get it should
    not cause an exception either."""
    item = Item.from_dict({"a": None})
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual_dict = adapter.get_field_meta("a")
    assert actual_dict == {}


def test_field_names():
    @attrs.define
    class _Item(Item):
        a: int
        b: Optional[int] = None

    item = _Item.from_dict({"a": 1, "c": 2, "d": None})
    with configured_adapter():
        adapter = ItemAdapter(item)
        actual = adapter.field_names()
    assert tuple(actual) == ("a", "b", "c", "d")


def test_known_field_get():
    url = "https://example.com/"
    product = Product(url=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter["url"] == url


def test_known_field_get_missing():
    url = "https://example.com/"
    product = Product(url=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter["canonicalUrl"] is None


def test_known_field_set():
    url = "https://example.com/"
    product = Product(url=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        adapter["canonicalUrl"] = url
        assert adapter["canonicalUrl"] == url
        assert product.canonicalUrl == url


def test_known_field_update():
    product = Product(url="https://example.com/a")
    with configured_adapter():
        adapter = ItemAdapter(product)
        new_url = "https://example.com/b"
        adapter["url"] = new_url
        assert adapter["url"] == new_url
        assert product.url == new_url


def test_known_field_remove():
    url = "https://example.com/"
    product = Product(url=url, canonicalUrl=url)
    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["canonicalUrl"]
        assert "canonicalUrl" not in adapter
        assert not hasattr(product, "canonicalUrl")


def test_known_field_remove_missing():
    product = Product(url="https://example.com/")
    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["canonicalUrl"]
        assert "canonicalUrl" not in adapter
        assert not hasattr(product, "canonicalUrl")


def test_known_field_remove_missing_twice():
    product = Product(url="https://example.com/")
    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["canonicalUrl"]
        with pytest.raises(KeyError):
            del adapter["canonicalUrl"]


def test_unknown_field_get():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating={"worstRating": 0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        assert adapter["a"] == "b"

        adapter = ItemAdapter(product.additionalProperties[0])
        assert adapter["max"] == 10

        adapter = ItemAdapter(product.aggregateRating)
        assert adapter["worstRating"] == 0


def test_unknown_field_get_missing():
    product = Product(url="https://example.com")
    with configured_adapter():
        adapter = ItemAdapter(product)
        with pytest.raises(KeyError):
            adapter["a"]


def test_unknown_field_set():
    product = Product.from_dict(
        dict(
            additionalProperties=[{"name": "a", "value": "b"}],
            aggregateRating={"bestRating": 5.0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        adapter["a"] = "b"
        assert adapter["a"] == "b"
        assert product._unknown_fields_dict["a"] == "b"

        adapter = ItemAdapter(product.additionalProperties[0])
        adapter["max"] = 10
        assert adapter["max"] == 10
        assert product.additionalProperties[0]._unknown_fields_dict["max"] == 10

        adapter = ItemAdapter(product.aggregateRating)
        adapter["worstRating"] = 0
        assert adapter["worstRating"] == 0
        assert product.aggregateRating._unknown_fields_dict["worstRating"] == 0


def test_unknown_field_update():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating={"worstRating": 0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        adapter["a"] = "c"
        assert adapter["a"] == "c"
        assert product._unknown_fields_dict["a"] == "c"

        adapter = ItemAdapter(product.additionalProperties[0])
        adapter["max"] = 20
        assert adapter["max"] == 20
        assert product.additionalProperties[0]._unknown_fields_dict["max"] == 20

        adapter = ItemAdapter(product.aggregateRating)
        adapter["worstRating"] = 1
        assert adapter["worstRating"] == 1
        assert product.aggregateRating._unknown_fields_dict["worstRating"] == 1


def test_unknown_field_remove():
    product = Product.from_dict(
        dict(
            a="b",
            additionalProperties=[{"name": "a", "value": "b", "max": 10}],
            aggregateRating={"worstRating": 0},
            url="https://example.com/",
        )
    )

    with configured_adapter():
        adapter = ItemAdapter(product)
        del adapter["a"]
        assert "a" not in adapter
        assert "a" not in product._unknown_fields_dict

        adapter = ItemAdapter(product.additionalProperties[0])
        del adapter["max"]
        assert "max" not in adapter
        assert "max" not in product.additionalProperties[0]._unknown_fields_dict

        adapter = ItemAdapter(product.aggregateRating)
        del adapter["worstRating"]
        assert "worstRating" not in adapter
        assert "worstRating" not in product.aggregateRating._unknown_fields_dict


def test_unknown_field_remove_missing():
    product = Product(url="https://example.com/")
    with configured_adapter():
        adapter = ItemAdapter(product)
        with pytest.raises(KeyError):
            del adapter["a"]


def test_keep_empty_adapter_global():
    @attrs.define
    class _Item(Item):
        children: CollectionType[Item]

    item = _Item([])
    with configured_adapter(ZyteItemKeepEmptyAdapter):
        adapter = ItemAdapter(item)
        actual_dict = adapter.asdict()
    assert actual_dict == {"children": []}


def test_keep_empty_adapter_local():
    @attrs.define
    class _Item(Item):
        children: CollectionType[Item]

    class TestAdapter(ItemAdapter):
        ADAPTER_CLASSES = (
            cast(Deque[Type[AdapterInterface]], deque([ZyteItemKeepEmptyAdapter]))
            + ItemAdapter.ADAPTER_CLASSES
        )

    item = _Item([])
    adapter = TestAdapter(item)
    actual_dict = adapter.asdict()
    assert actual_dict == {"children": []}



import html
from datetime import datetime
from typing import Any, Generic, Optional, Type, TypeVar, Union

import attrs
import html_text
from clear_html import cleaned_node_to_text
from lxml.html import HtmlElement
from price_parser import Price
from web_poet import ItemPage, RequestUrl, Returns, WebPage, field
from web_poet.fields import FieldsMixin
from web_poet.pages import ItemT
from web_poet.utils import ensure_awaitable, get_generic_param

from .components import (
    ArticleListMetadata,
    ArticleMetadata,
    ArticleNavigationMetadata,
    BusinessPlaceMetadata,
    JobPostingMetadata,
    ProductListMetadata,
    ProductMetadata,
    ProductNavigationMetadata,
    RealEstateMetadata,
    request_list_processor,
)
from .items import (
    Article,
    ArticleList,
    ArticleNavigation,
    BusinessPlace,
    JobPosting,
    Product,
    ProductList,
    ProductNavigation,
    RealEstate,
)
from .processors import (
    brand_processor,
    breadcrumbs_processor,
    description_html_processor,
    description_processor,
    price_processor,
    simple_price_processor,
)
from .util import format_datetime, metadata_processor

#: Generic type for metadata classes for specific item types.
MetadataT = TypeVar("MetadataT")


def _date_downloaded_now():
    return format_datetime(datetime.utcnow())


class HasMetadata(Generic[MetadataT]):
    """Inherit from this generic mixin to set the metadata class used by a page
    class."""

    @property
    def metadata_cls(self) -> Optional[Type[MetadataT]]:
        """Metadata class."""
        return _get_metadata_class(type(self))


def _get_metadata_class(cls: type) -> Optional[Type[MetadataT]]:
    return get_generic_param(cls, HasMetadata)


class PriceMixin(FieldsMixin):
    """Provides price-related field implementations."""

    _parsed_price: Optional[Price] = None

    async def _get_parsed_price(self) -> Optional[Price]:
        if self._parsed_price is None:
            # the price field wasn't executed or doesn't write _parsed_price
            price = getattr(self, "price", None)
            price = await ensure_awaitable(price)
            if self._parsed_price is None:
                # the price field doesn't write _parsed_price (or doesn't exist)
                self._parsed_price = Price(
                    amount=None, currency=None, amount_text=price
                )
        return self._parsed_price

    @field
    def currency(self) -> Optional[str]:
        return getattr(self, "CURRENCY", None)

    @field
    async def currencyRaw(self) -> Optional[str]:
        parsed_price = await self._get_parsed_price()
        if parsed_price:
            return parsed_price.currency
        return None


class DescriptionMixin(FieldsMixin):
    """Provides description and descriptionHtml field implementations."""

    UNSET = object()

    _descriptionHtml_node: Any = UNSET
    _description_node: Any = UNSET
    _description_str: Any = UNSET

    _description_default = False
    _descriptionHtml_default = False

    @staticmethod
    def wrap_description_into_html(description: str) -> str:
        r"""Convert plain text into an article HTML.

        The format tries to match clear_html.cleaned_node_to_html().

        >>> DescriptionMixin.wrap_description_into_html('')
        '<article>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo')
        '<article>\n\n<p>foo</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo\nbar')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo\n\nbar')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('\nfoo\n\nbar\n')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo\nbar\n\nbaz\n')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n<p>baz</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('2>1')
        '<article>\n\n<p>2&gt;1</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('<p>')
        '<article>\n\n<p>&lt;p&gt;</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('&lt;p&gt;')
        '<article>\n\n<p>&amp;lt;p&amp;gt;</p>\n\n</article>'
        """
        paras_wrapped = [
            f"\n<p>{html.escape(para)}</p>\n"
            for para in description.split("\n")
            if para
        ]
        return f"<article>\n{''.join(paras_wrapped)}\n</article>"

    async def _get_description(self) -> Optional[str]:
        if self._description_default:
            return None
        if self._description_str == self.UNSET:
            description = await ensure_awaitable(self.description)
            if self._description_str == self.UNSET:
                # the description field doesn't write _description_str
                self._description_str = description
        return self._description_str

    async def _get_description_html(self) -> Optional[HtmlElement]:
        if self._descriptionHtml_default:
            return None
        if self._descriptionHtml_node == self.UNSET:
            descriptionHtml = await ensure_awaitable(self.descriptionHtml)
            if self._descriptionHtml_node == self.UNSET:
                # the descriptionHtml field doesn't write _descriptionHtml_node
                self._descriptionHtml_node = descriptionHtml
        return self._descriptionHtml_node

    @field
    async def description(self) -> Optional[str]:
        self._description_default = True
        description_html = await self._get_description_html()
        if isinstance(description_html, HtmlElement):
            return cleaned_node_to_text(description_html)
        if isinstance(description_html, str):
            return html_text.extract_text(description_html)
        return None

    @field
    async def descriptionHtml(self) -> Union[HtmlElement, str, None]:
        self._descriptionHtml_default = True
        description = await self._get_description()
        if self._description_node not in {self.UNSET, None}:
            # we can use the element provided by the description field
            return self._description_node
        if isinstance(description, str):
            return self.wrap_description_into_html(description)
        return None


class _BasePage(ItemPage[ItemT], HasMetadata[MetadataT]):
    class Processors:
        metadata = [metadata_processor]

    @field
    def metadata(self) -> MetadataT:
        if self.metadata_cls is None:
            raise ValueError(f"{type(self)} doesn'have a metadata class configured.")
        value = self.metadata_cls()
        attributes = dir(value)
        if "dateDownloaded" in attributes:
            value.dateDownloaded = _date_downloaded_now()  # type: ignore
        if "probability" in attributes:
            value.probability = 1.0  # type: ignore
        return value

    def no_item_found(self) -> ItemT:
        """Return an item with the current url and probability=0,
        indicating that the passed URL doesn't contain the expected item.

        Use it in your .validate_input implementation.
        """
        if self.metadata_cls is None:
            raise ValueError(f"{type(self)} doesn'have a metadata class configured.")
        metadata = self.metadata_cls()
        metadata_attributes = dir(metadata)
        if "dateDownloaded" in metadata_attributes:
            metadata.dateDownloaded = _date_downloaded_now()  # type: ignore
        if "probability" in metadata_attributes:
            metadata.probability = 0.0  # type: ignore
        return self.item_cls(  # type: ignore
            url=self.url,  # type: ignore[attr-defined]
            metadata=metadata,
        )


@attrs.define
class BasePage(_BasePage):
    class Processors(_BasePage.Processors):
        pass

    request_url: RequestUrl

    @field
    def url(self) -> str:
        return str(self.request_url)


class BaseArticlePage(BasePage, Returns[Article], HasMetadata[ArticleMetadata]):
    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]


class BaseArticleListPage(
    BasePage, Returns[ArticleList], HasMetadata[ArticleListMetadata]
):
    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]


class BaseArticleNavigationPage(
    BasePage, Returns[ArticleNavigation], HasMetadata[ArticleNavigationMetadata]
):
    pass


class BaseBusinessPlacePage(
    BasePage, Returns[BusinessPlace], HasMetadata[BusinessPlaceMetadata]
):
    class Processors(BasePage.Processors):
        description = [description_processor]


class BaseJobPostingPage(
    BasePage, DescriptionMixin, Returns[JobPosting], HasMetadata[JobPostingMetadata]
):
    class Processors(BasePage.Processors):
        description = [description_processor]
        descriptionHtml = [description_html_processor]


class BaseProductPage(
    BasePage,
    DescriptionMixin,
    PriceMixin,
    Returns[Product],
    HasMetadata[ProductMetadata],
):
    class Processors(BasePage.Processors):
        brand = [brand_processor]
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]
        descriptionHtml = [description_html_processor]
        price = [price_processor]
        regularPrice = [simple_price_processor]


class BaseProductListPage(
    BasePage, Returns[ProductList], HasMetadata[ProductListMetadata]
):
    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]


class BaseProductNavigationPage(
    BasePage, Returns[ProductNavigation], HasMetadata[ProductNavigationMetadata]
):
    class Processors(BasePage.Processors):
        subCategories = [request_list_processor]
        items = [request_list_processor]


class BaseRealEstatePage(
    BasePage, Returns[RealEstate], HasMetadata[RealEstateMetadata]
):
    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]


@attrs.define
class Page(_BasePage, WebPage):
    class Processors(_BasePage.Processors):
        pass

    @field
    def url(self) -> str:
        return str(self.response.url)


class ArticlePage(Page, Returns[Article], HasMetadata[ArticleMetadata]):
    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]


class ArticleListPage(Page, Returns[ArticleList], HasMetadata[ArticleListMetadata]):
    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]


class ArticleNavigationPage(
    Page, Returns[ArticleNavigation], HasMetadata[ArticleNavigationMetadata]
):
    pass


class BusinessPlacePage(
    Page, Returns[BusinessPlace], HasMetadata[BusinessPlaceMetadata]
):
    class Processors(Page.Processors):
        description = [description_processor]


class JobPostingPage(
    Page, DescriptionMixin, Returns[JobPosting], HasMetadata[JobPostingMetadata]
):
    class Processors(Page.Processors):
        description = [description_processor]
        descriptionHtml = [description_html_processor]


class ProductPage(
    Page, DescriptionMixin, PriceMixin, Returns[Product], HasMetadata[ProductMetadata]
):
    class Processors(Page.Processors):
        brand = [brand_processor]
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]
        descriptionHtml = [description_html_processor]
        price = [price_processor]
        regularPrice = [simple_price_processor]


class ProductListPage(Page, Returns[ProductList], HasMetadata[ProductListMetadata]):
    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]


class ProductNavigationPage(
    Page, Returns[ProductNavigation], HasMetadata[ProductNavigationMetadata]
):
    pass


class RealEstatePage(Page, Returns[RealEstate], HasMetadata[RealEstateMetadata]):
    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]



from os.path import dirname, join

from setuptools import find_packages, setup

with open(join(dirname(__file__), "zyte_common_items/VERSION"), "rb") as f:
    version = f.read().decode("ascii").strip()

setup(
    name="zyte-common-items",
    version=version,
    description="Item definitions for Zyte API schema",
    long_description=open("README.rst").read(),
    long_description_content_type="text/x-rst",
    author="Zyte Group Ltd",
    author_email="info@zyte.com",
    url="https://github.com/zytedata/zyte-common-items",
    packages=find_packages(
        exclude=[
            "tests",
        ]
    ),
    package_data={
        "zyte_common_items": ["py.typed", "VERSION"],
    },
    # needs to be in sync with the tox.ini min env
    install_requires=[
        "attrs>=22.1.0",
        "clear-html>=0.4.0",
        "itemadapter>=0.8.0",
        "price-parser>=0.3.4",
        "web-poet>=0.14.0",
        "zyte-parsers>=0.3.0",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    python_requires=">=3.8",
)



from collections.abc import Iterable
from functools import wraps
from typing import Any, Callable, List, Optional, Union

from clear_html import clean_node, cleaned_node_to_html, cleaned_node_to_text
from lxml.html import HtmlElement
from parsel import Selector, SelectorList
from price_parser import Price
from web_poet.mixins import ResponseShortcutsMixin
from zyte_parsers import Breadcrumb as zp_Breadcrumb
from zyte_parsers import extract_brand_name, extract_breadcrumbs, extract_price

from .items import Breadcrumb


def _get_base_url(page: Any) -> Optional[str]:
    if isinstance(page, ResponseShortcutsMixin):
        return page.base_url
    return getattr(page, "url", None)


def _handle_selectorlist(value: Any) -> Any:
    if not isinstance(value, SelectorList):
        return value
    if len(value) == 0:
        return None
    return value[0]


def _format_price(price: Price) -> Optional[str]:
    """Return the price amount as a string, with a minimum of 2 decimal
    places."""
    if price.amount is None:
        return None
    *_, exponent = price.amount.as_tuple()
    if not isinstance(exponent, int):
        return None  # NaN, Infinity, etc.
    if exponent <= -2:
        return str(price.amount)
    return f"{price.amount:.2f}"


def only_handle_nodes(
    f: Callable[[Union[Selector, HtmlElement], Any], Any]
) -> Callable[[Any, Any], Any]:
    @wraps(f)
    def wrapper(value: Any, page: Any) -> Any:
        value = _handle_selectorlist(value)
        if not isinstance(value, (Selector, HtmlElement)):
            return value
        result = f(value, page)
        return result

    return wrapper


def breadcrumbs_processor(value: Any, page: Any) -> Any:
    """Convert the data into a list of :class:`~zyte_common_items.Breadcrumb` objects if possible.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList`, :class:`~lxml.html.HtmlElement` and
    an iterable of :class:`zyte_parsers.Breadcrumb` objects. Other inputs are
    returned as is.
    """

    def _from_zp_breadcrumb(value: zp_Breadcrumb) -> Breadcrumb:
        return Breadcrumb(name=value.name, url=value.url)

    value = _handle_selectorlist(value)

    if isinstance(value, (Selector, HtmlElement)):
        zp_breadcrumbs = extract_breadcrumbs(value, base_url=_get_base_url(page))
        return (
            [_from_zp_breadcrumb(b) for b in zp_breadcrumbs] if zp_breadcrumbs else None
        )

    if not isinstance(value, Iterable) or isinstance(value, str):
        return value

    results: List[Any] = []
    for item in value:
        if isinstance(item, zp_Breadcrumb):
            results.append(_from_zp_breadcrumb(item))
        else:
            results.append(item)
    return results


@only_handle_nodes
def brand_processor(value: Union[Selector, HtmlElement], page: Any) -> Any:
    """Convert the data into a brand name if possible.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.
    """
    return extract_brand_name(value, search_depth=2)


@only_handle_nodes
def price_processor(value: Union[Selector, HtmlElement], page: Any) -> Any:
    """Convert the data into a price string if possible.

    Uses the price-parser_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.

    Puts the parsed Price object into ``page._parsed_price``.

    .. _price-parser: https://github.com/scrapinghub/price-parser
    """
    price = extract_price(value)
    page._parsed_price = price
    return _format_price(price)


@only_handle_nodes
def simple_price_processor(value: Union[Selector, HtmlElement], page: Any) -> Any:
    """Convert the data into a price string if possible.

    Uses the price-parser_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.

    .. _price-parser: https://github.com/scrapinghub/price-parser
    """
    price = extract_price(value)
    return _format_price(price)


@only_handle_nodes
def description_html_processor(value: Union[Selector, HtmlElement], page: Any) -> Any:
    """Convert the data into a cleaned up HTML if possible.

    Uses the clear-html_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.

    Puts the cleaned HtmlElement object into ``page._descriptionHtml_node``.

    .. _clear-html: https://github.com/zytedata/clear-html
    """
    if isinstance(value, Selector):
        value = value.root
    if value is None:
        return None
    assert isinstance(value, HtmlElement)
    cleaned_node = clean_node(value, _get_base_url(page))
    page._descriptionHtml_node = cleaned_node
    return cleaned_node_to_html(cleaned_node)


def description_processor(value: Any, page: Any) -> Any:
    """Convert the data into a cleaned up text if possible.

    Uses the clear-html_ library.

    Supported inputs are :class:`~parsel.selector.Selector`,
    :class:`~parsel.selector.SelectorList` and :class:`~lxml.html.HtmlElement`.
    Other inputs are returned as is.

    Puts the cleaned HtmlElement object into ``page._description_node`` and the
    cleaned text into ``page._description_str``.

    .. _clear-html: https://github.com/zytedata/clear-html
    """
    value = _handle_selectorlist(value)
    if isinstance(value, str):
        page._description_str = value
        return value
    if isinstance(value, Selector):
        value = value.root
    if value is None:
        return None
    assert isinstance(value, HtmlElement)
    cleaned_node = clean_node(value, _get_base_url(page))
    cleaned_text = cleaned_node_to_text(cleaned_node)
    page._description_node = cleaned_node
    page._description_str = cleaned_text
    return cleaned_text



from typing import Any, Callable, Dict, Optional, Tuple, Type, Union
from warnings import warn
from weakref import WeakKeyDictionary

import attrs
from web_poet.page_inputs.url import _Url

# Caches the attribute names for attr.s classes
CLASS_ATTRS: WeakKeyDictionary = WeakKeyDictionary()


def split_in_unknown_and_known_fields(
    data: Optional[dict], item_cls: Type
) -> Tuple[Dict, Dict]:
    """
    Return a pair of dicts. The first one contains those elements not belonging to the
    attr class ``item_cls``. The second one contains the rest. That is, those
    attributes not belonging to ``item_cls`` class
    """
    data = data or {}
    if not attrs.has(item_cls):
        raise ValueError(f"The cls {item_cls} is not attrs class")
    if item_cls not in CLASS_ATTRS:
        CLASS_ATTRS[item_cls] = {field.name for field in attrs.fields(item_cls)}
    unknown, known = split_dict(data, lambda k: k in CLASS_ATTRS[item_cls])
    return unknown, known


def split_dict(dict: Dict, key_pred: Callable[[Any], Any]) -> Tuple[Dict, Dict]:
    """Splits the dictionary in two.

    The first dict contains the records for which the key predicate is False
    and the second dict contains the rest.

    >>> split_dict({}, lambda k: False)
    ({}, {})
    >>> split_dict(dict(a=1, b=2, c=3), lambda k: k != 'a')
    ({'a': 1}, {'b': 2, 'c': 3})
    """  # noqa
    yes, no = {}, {}
    for k, v in dict.items():
        if key_pred(k):
            yes[k] = v
        else:
            no[k] = v
    return (no, yes)


def url_to_str(url: Union[str, _Url]) -> str:
    if not isinstance(url, (str, _Url)):
        raise ValueError(
            f"{url!r} is neither a string nor an instance of RequestURL or ResponseURL."
        )
    return str(url)


def format_datetime(dt):
    return f"{dt.isoformat(timespec='seconds')}Z"


def convert_to_class(value: Any, new_cls: type) -> Any:
    if type(value) == new_cls:
        return value
    input_attributes = {attribute.name for attribute in attrs.fields(value.__class__)}
    output_attributes = {attribute.name for attribute in attrs.fields(new_cls)}
    shared_attributes = input_attributes & output_attributes
    new_value = new_cls(
        **{attribute: getattr(value, attribute) for attribute in shared_attributes}
    )
    removed_nonempty_attributes = {
        attribute
        for attribute in (input_attributes - output_attributes)
        if getattr(value, attribute)
        != attrs.fields_dict(value.__class__)[attribute].default
    }
    if removed_nonempty_attributes:
        warn(
            (
                f"Conversion of {value} into {new_cls} is dropping the non-default "
                f"values of the following attributes: "
                f"{removed_nonempty_attributes}."
            ),
            RuntimeWarning,
        )
    return new_value


def cast_metadata(value, cls):
    new_value = convert_to_class(value, cls)
    return new_value


def metadata_processor(metadata, page):
    return cast_metadata(metadata, page.metadata_cls)


class MetadataCaster:
    def __init__(self, target):
        self._target = target

    def __call__(self, value):
        return cast_metadata(value, self._target)
